#' Identifies and connects dangles in network generated by Network Nav function
#' @param inCOM  list list of input COMIDs
#' @param nhdDF sf data.frame (data frame) valid data frame of NHD flowlines
#' @import dplyr
#' @return (list) list of COMIDs connecting dangle to existing network
#' @export
#' @examples
#' source(system.file("extdata", "sample_flines.R", package = "nhdplusTools"))
#'
#' sample_flines <- nhdplusTools::prepare_nhdplus(sample_flines, 0, 0, 0, FALSE) %>%
#'   left_join(select(sample_flines, COMID, DnHydroseq, DnLevelPat, Pathlength))
#'
#' comid <- sample(sample_flines$COMID, 10)
#'
#' NetworkConnection(comid, sample_flines)

NetworkConnection <- function(incom, nhd, status = FALSE){

  nhd <- dplyr::group_by(nhd, LevelPathI) %>%
    dplyr::mutate(DnLevelPat = min(DnLevelPat)) %>%
    dplyr::ungroup()

  upnet_DF <- filter(nhd, COMID %in% incom) %>%
    filter(!DnHydroseq %in% Hydroseq)

  # while the number of dangles is greater than 0
  while (length(upnet_DF$COMID) > 0) {

    # create item for number of dangles
    count <- dim(upnet_DF)[1]

    if(status)
      message(dim(upnet_DF))

    # find out which level paths are downstream of dangling huc12 POIs
    DSLP <- upnet_DF %>% pull(DnLevelPat)#[upnet_DF$COMID %in% incom]
    CLP <- upnet_DF %>% pull(LevelPathI)

    # Get the COMID of the hydroseq with level path value
    # the lowest downstream flowline within the levelpath
    inCom2 <- nhd$COMID[nhd$Hydroseq %in% c(DSLP, CLP)]

    # Run the upstream navigation code

    upNet <- unique(unlist(lapply(inCom2, function(x, nhd) {
      nhdplusTools::get_UM(nhd, x, include = TRUE)
    }, nhd = nhd)))

    # Append result to existing segment list
    incom <- append(incom, upNet)

    # Get the same variable as above
    upnet_DF <- filter(nhd, COMID %in% incom, !DnHydroseq %in% Hydroseq)

    # Get the count
    count2 <- dim(upnet_DF)[1]

    # if the count has remained the same we are done and return the flowline list
    if (count == count2){
      return (incom)
    }
  }

  return(incom)
}
