#' get_sb_file
#' @param item sciencebase item id
#' @param item_files character vector of file names. "all" will download all files.
#' @param out_destination character path to save files into
#' @param unzip logical if TRUE, files ending with .7z will be un7zipped
#' @param check_path character path to check, if it exists, it will be returned invisibly
#' @return path that data was saved to
get_sb_file <- function(item, item_files, out_destination, unzip = TRUE, check_path = NULL) {
  
  if(!is.null(check_path) && file.exists(check_path)) return(invisible(check_path))
  
  check_auth()
  
  if(length(item_files) == 1 && item_files == "all") {
    item_files <- sbtools::item_list_files(item)$fname
  }
  
  if(!dir.exists(out_destination)) dir.create(out_destination, showWarnings = FALSE, recursive = TRUE)
  
  out_files <- file.path(out_destination, item_files)
  
  missing <- out_files[!file.exists(out_files)]
  
  if(length(missing) > 0) {
    sbtools::item_file_download(item, 
                                names = basename(missing), 
                                destinations = missing)
  }
  
  if(unzip) {
    
    un7zip_fs <- missing[grepl("7z$|zip$", missing)]

    for(f in un7zip_fs) {
      system(paste0(sevenz, " e -o", out_destination, " ", f))
    }
    
  }
  
  if(!is.null(check_path) && file.exists(check_path)) return(invisible(check_path))
  
  if(!is.null(check_path))
    stop("did not create ", check_path)
  
  invisible(out_destination)
  
}

#' download a file
#' @param url character url where the file can be retrieved
#' @param out_path character path the directory to save the downloaded file
#' @param check_path character path to check that should exist when done
#' @param file_name character name to use instead of basename of url
#' @param unzip logical whether to try to unzip the file
download_file <- function(url, out_path, check_path = NULL, unzip = TRUE, file_name = NULL) {

  if(!is.null(check_path) && file.exists(check_path)) return(invisible(check_path))

  dir.create(out_path, showWarnings = FALSE, recursive = TRUE)
  
  if(is.null(file_name)) file_name <- basename(url)
  
  out <- file.path(out_path, file_name)
  
  if(!file.exists(out)) {
    httr::RETRY("GET", url, httr::write_disk(out))
  
    if(unzip & grepl("7z$|zip$", out))
      system(paste0(sevenz, " x ", out, " -o", out_path))
    
  }
  
  if(!is.null(check_path) && file.exists(check_path)) return(invisible(check_path))
  
  if(!is.null(check_path))
    stop("did not create ", check_path)
}

#' make flow direction and flow accumulation file list
#' @param fdr_fac_dir directory where flow direction and flow accumulation are
#' @param hu2 two digit huc to pass to download_elev
#' @return list containing all flow direction and flow accumulation files
make_fdr_fac_list <- function(fdr_fac_dir, hu2) {
  if(!dir.exists(fdr_fac_dir))
    download_elev("FDRFAC", fdr_fac_dir)
  
  dirs <- unique(dirname(list.files(fdr_fac_dir, recursive = TRUE, full.names = TRUE)))
  
  fdr <- dirs[grepl(".*/fdr$", dirs)]
  fac <- dirs[grepl(".*/fac$", dirs)]
  
  out <- list(fdr = list(), fac = list())
  
  rpu <- substr(fdr, (nchar(fdr) - 6), (nchar(fdr) - 4))
  
  out$fdr <- as.list(setNames(fdr, paste0("rpu_", rpu)))
  out$fac <- as.list(setNames(fac, paste0("rpu_", rpu)))
  
  out
}

make_nhdplus_elev_list <- function(elev_dir, region, hu2) {
  
  if(!dir.exists(elev_dir))
    download_elev("DEM", elev_dir, hu2)
  
  dirs <- unique(dirname(list.files(elev_dir, recursive = TRUE, full.names = TRUE)))
  
  elev_cm <- dirs[grepl(".*/elev_cm$", dirs)]
  
  out <- list(elev_cm = list())
  
  rpu <- substr(elev_cm, (nchar(elev_cm) - 10), (nchar(elev_cm) - 8))
  
  out$elev_cm <- as.list(setNames(elev_cm, paste0("rpu_", rpu)))
  
  out
  
}
