#' get highest drainage area
#' @param lookup lookup table from refactor step
#' @param flowline source flowlines of refactor
#' @param refactored refactored but not reconciled reconciled_ID and hydrosequence values
#' 
get_highest_da <- function(lookup, flowline, refactored) {
  lookup |>
    inner_join(select(st_drop_geometry(flowline), COMID, TotDASqKM),
               by = c("NHDPlusV2_COMID" = "COMID")) |>
    inner_join(refactored, by = c("NHDPlusV2_COMID" = "COMID")) |>
    group_by(reconciled_ID) |>
    filter(Hydroseq == min(Hydroseq))
}

#' aggregate catchments function
#' @param divides reconciled divides from refactor step
#' @param reconciled refactored and reconciled flowlines from refactor step
#' @param outlets_POI POI outlets from refactor step
#' @param aggregate_da_thresh_sqkm drainage area threshold passed to aggregate catchments
#' @param cache_split rda file to output error information to
#' @param highest_DA as generated by get_highest_da()
#' 
aggregate_catchments_fun <- function(divides, reconciled, outlets_POI, 
                                     aggregate_da_thresh_sqkm, cache_split,
                                     highest_DA) {
  
  # https://github.com/r-spatial/sf/issues/1094#issuecomment-988933580 for why this is here.
  st_geometry(divides[st_is_empty(divides),]) <-
    st_cast(st_geometry(divides[st_is_empty(divides),]), "MULTIPOLYGON")
  
  # POI outlets
  outlets <- select(st_drop_geometry(outlets_POI), ID = reconciled_ID, type)
  
  # Identify reconciled flowpaths that are terminal and above DA threshold
  reconciled_DA <- filter(reconciled, TotDASqKM > aggregate_da_thresh_sqkm, is.na(toID))
  
  # Ones not in existing pois
  missing_outlets <- filter(reconciled_DA, !ID %in% outlets$ID) |>
    pull(ID)
  
  # bind missing outlets, re-assign mis-attributed terminals
  if(length(missing_outlets) > 0){
    outlets <- bind_rows(outlets,
                         data.frame(ID = missing_outlets,
                                    type = rep("terminal", length(missing_outlets))))
    # write_sf(filter(reconciled, ID %in% missing_outlets), out_agg_gpkg, "missing_outlets")
  }
  
  # Identify reconciled flowpaths that are terminal and below DA threshold
  reconciled_DA <- filter(reconciled, TotDASqKM <= aggregate_da_thresh_sqkm, is.na(toID))
  
  # Ones not in existing pois
  extra_outlets <- filter(reconciled_DA, !ID %in% outlets$ID) |>
    pull(ID)
  
  extra_nets <- hydroloom::sort_network(select(st_drop_geometry(reconciled),
                                               ID, toID), outlets = extra_outlets)
  
  # remove extra network from contention
  reconciled <- filter(reconciled, !ID %in% extra_nets$ID)
  
  outlets <- outlets |>
    left_join(select(st_drop_geometry(reconciled), ID, toID), by = "ID") |>
    mutate(type = ifelse(is.na(toID), "terminal", "outlet")) |>
    mutate(type = ifelse(type == "terminal" & !is.na(toID), "outlet", type)) |>
    select(-toID)
  
  outlets <- filter(outlets, ID %in% reconciled$ID) |> 
    distinct()
  
  agg_cats <- hyRefactor::aggregate_catchments(flowpath = reconciled,
                                   divide = divides,
                                   outlets = outlets,
                                   da_thresh = aggregate_da_thresh_sqkm,
                                   only_larger = TRUE,
                                   post_mortem_file = cache_split)
  
  agg_cats$cat_sets$set <- sapply(agg_cats$cat_sets$set, paste, collapse = ",")
  agg_cats$fline_sets$set <- sapply(agg_cats$fline_sets$set, paste, collapse = ",")
  
  # write_sf(agg_cats$cat_sets, out_agg_gpkg, agg_cats_layer)
  # write_sf(agg_cats$fline_sets, out_agg_gpkg, agg_fline_layer)
  
  # Get physical geometry  of reconciled FL end node whose ID is an aggregated outlet
  rec_outlets <- filter(st_drop_geometry(reconciled), ID %in% agg_cats$fline_sets$ID) |>
    cbind(hydroloom::get_node(filter(reconciled, ID %in% agg_cats$fline_sets$ID))) |>
    st_as_sf()
  
  # Build final POI set
  POIs_att <- select(rec_outlets, ID, toID) |> 
    # Bring over POI information
    inner_join(st_drop_geometry(outlets_POI), by = c("ID" = "reconciled_ID")) |>
    select(-c(type, member_COMID))
  
  POIs_missing <- filter(select(rec_outlets, -toID), !ID %in% POIs_att$ID) |>
    inner_join(st_drop_geometry(agg_cats$fline_sets), by = "ID") |>
    arrange(ID) |>
    mutate(poi_id = max(POIs_att$poi_id) + row_number()) |>
    select(ID, toID, poi_id, event_id = event_identifier, Type_Term = orig_levelpathID) |>
    inner_join(select(highest_DA, NHDPlusV2_COMID, reconciled_ID), by = c("ID" = "reconciled_ID")) |>
    select(ID, toID, hy_id = NHDPlusV2_COMID, poi_id, event_id, Type_Term) |>
    mutate(Type_Term = ifelse(is.na(toID), Type_Term, NA),
           Type_Con = ifelse(!is.na(toID), 1, NA)) |>
    st_as_sf()
  
  if(nrow(POIs_missing) > 0){
    final_POIs <- data.table::rbindlist(list(POIs_att, select(POIs_missing, 
                                                              -c(Type_Term, Type_Con))), fill = TRUE) |>
      st_as_sf()
  } else {
    final_POIs <- POIs_att
  }
  
  unaggregated_outlets <- filter(outlets_POI, !reconciled_ID %in% final_POIs$ID)
  double_outlets <- final_POIs |> group_by(ID) |> filter(n() > 1)
  
  # Write out
  # write_sf(final_POIs, out_agg_gpkg, mapped_outlets_layer)
  # write_sf(unaggregated_outlets, out_agg_gpkg, "unmapped_outlets")
  # write_sf(double_outlets, out_agg_gpkg, "double_outlets")
  
  list(mapped_outlets_layer = final_POIs, unmapped_outlets = unaggregated_outlets,
       double_outlets = double_outlets, missing_outlets = filter(reconciled, ID %in% missing_outlets),
       agg_cats_layer = agg_cats$cat_sets, agg_fline_layer = agg_cats$fline_sets)
  
}

#' make long form pois
#' @param final_POIs final points of interest from aggregate step
#' 
make_long_form_pois <- function(final_POIs) {
  
  POIs <- final_POIs |>
    arrange(ID) 
  
  final_POI_geom <- POIs |>
    select(ID)
  
  final_POI_geom <- final_POI_geom |>
    cbind(st_coordinates(final_POI_geom)) |>
    group_by(ID, X, Y) |>
    mutate(geom_id = cur_group_id()) |>
    ungroup()
  
  POIs |>
    inner_join(select(st_drop_geometry(final_POI_geom), -X, -Y), by = "ID") |>
    distinct()
  
}

#' make lookup
#' @param reconciled refactored and reconciled flowlines from refactor step
#' @param agg_cats aggregated catchments from aggregate step
#' @param agg_fline aggregated flowlines from aggregate step
#' 
make_lookup <- function(reconciled, agg_cats, agg_fline) {
  
  refactor_lookup <- dplyr::select(st_drop_geometry(reconciled), ID, member_COMID) |>
    dplyr::mutate(member_COMID = strsplit(member_COMID, ",")) |>
    tidyr::unnest(cols = member_COMID) |>
    dplyr::mutate(NHDPlusV2_COMID = as.integer(member_COMID)) |> # note as.integer truncates
    dplyr::rename(reconciled_ID = ID)
  
  aggregate_lookup_fline <- dplyr::select(st_drop_geometry(agg_fline), ID, set) |>
    dplyr::mutate(set = strsplit(set, ",")) |>
    tidyr::unnest(cols = set) |>
    dplyr::rename(aggregated_flowpath_ID = ID, reconciled_ID = set)
  
  aggregate_lookup_catchment <- dplyr::select(st_drop_geometry(agg_cats), ID, set) |>
    dplyr::mutate(set = strsplit(set, ",")) |>
    tidyr::unnest(cols = set) |>
    dplyr::rename(aggregated_divide_ID = ID, reconciled_ID = set)
  
  if(is.character(aggregate_lookup_catchment$reconciled_ID)) 
    aggregate_lookup_catchment$reconciled_ID <- as.integer(aggregate_lookup_catchment$reconciled_ID)
  if(is.character(aggregate_lookup_fline$reconciled_ID)) 
    aggregate_lookup_fline$reconciled_ID <- as.integer(aggregate_lookup_fline$reconciled_ID)
  
  tibble::tibble(NHDPlusV2_COMID = unique(as.integer(refactor_lookup$member_COMID))) |>
    dplyr::left_join(refactor_lookup, by = "NHDPlusV2_COMID") |>
    dplyr::left_join(aggregate_lookup_fline, by = "reconciled_ID") |>
    dplyr::left_join(aggregate_lookup_catchment, by = "reconciled_ID")
  
}
