#'merge rpu to vpu
#' @param pois pois from reference fabric step
#' @param rpu_vpu table of rpu and vpu codes
#' @param vpu vpu to be merged
#' @param rpu_vpu_out outputs from rpus to other rpus/vpus
#' @param ref_gpkg_fun function to convert a rpu into a refactored gpkg path
#' @param agg_gpkg_fun function to convert a rpu into an aggregate gpkg path
#' @param lookup_table_refactor table name to get the lookup table from
#' @param reconciled_layer table name to get the reconciled flowlines from
#' @param divide_layer table name to get the reconciled divides from
#' @param split_divide_layer table name to get the split catchment divides from
#' @param agg_fline_layer table name to get the aggregate flowlines from
#' @param agg_cats_layer table name to get the aggregate catchments from
#' @param mapped_oulets_layer table name to get the long form mapped pois from
#' 
merge_rpu_to_vpu <- function(pois, rpu_vpu, vpu, rpu_vpu_out,
                             ref_gpkg_fun, agg_gpkg_fun, 
                             lookup_table_refactor,
                             reconciled_layer, divide_layer, split_divide_layer,
                             agg_fline_layer, agg_cats_layer, mapped_outlets_layer) {
  
  rpu_vpu <- rpu_vpu[rpu_vpu$vpuid == vpu,]
  
  # Thematic POIs
  POIs <- pois %>% 
    select(-geom_id)
  
  merged_layers <- merge_refactor(rpu_vpu$rpuid, 
                                  rpu_vpu_out, 
                                  ref_gpkg_fun,
                                  agg_gpkg_fun,
                                  lookup_table_refactor, 
                                  reconciled_layer, 
                                  divide_layer, 
                                  split_divide_layer,
                                  agg_fline_layer,
                                  agg_cats_layer,
                                  mapped_outlets_layer)
  
  merged_layers[[agg_cats_layer]] <- merged_layers[[agg_cats_layer]] %>%
    mutate(areasqkm = as.numeric(units::set_units(st_area(.), "km^2")))
  
  merged_layers[[agg_fline_layer]] <- merged_layers[[agg_fline_layer]] %>%
    mutate(lengthkm = as.numeric(units::set_units(st_length(.), "km"))) %>%
    left_join(select(st_drop_geometry(merged_layers[[agg_cats_layer]]), ID, areasqkm), by = "ID")
  
  cat_net_rfc <- st_drop_geometry(merged_layers[[reconciled_layer]]) %>%
    select(id = ID, toID = toID, lengthkm = LENGTHKM, levelpathid = LevelPathID) %>%
    left_join(select(st_drop_geometry(merged_layers[[divide_layer]]), 
                     id = ID, areasqkm), by = "id") %>%
    relocate(id, toID, lengthkm, areasqkm, levelpathid) 
  
  cat_net_gf <- st_drop_geometry(merged_layers[[agg_fline_layer]]) %>%
    select(id = ID, toid = toID, lengthkm, areasqkm) %>%
    left_join(distinct(select(merged_layers[[lookup_table_refactor]], 
                              id = aggregated_flowpath_ID, 
                              levelpathid = LevelPathID))) 
  
  list(pois = POIs, 
       lookup_table_refactor = select(merged_layers[[lookup_table_refactor]], 
                                      -aggregated_flowpath_ID, 
                                      -aggregated_divide_ID),
       reconciled_layer = merged_layers[[reconciled_layer]],
       divide_layer = merged_layers[[divide_layer]],
       split_divide_layer = merged_layers[[split_divide_layer]],
       mapped_outlets_layer = merged_layers[[mapped_outlets_layer]],
       agg_cats_layer = merged_layers[[agg_cats_layer]],
       agg_fline_layer = merged_layers[[agg_fline_layer]],
       lookup_table_refactor = merged_layers[[lookup_table_refactor]],
       reconciled_layer = merged_layers[[reconciled_layer]],
       catchment_network_table_refector = cat_net_rfc,
       cathment_network_table_aggregate = cat_net_gf,
       lookup_table_aggregate = merged_layers[[lookup_table_refactor]])
}

#' merge refactor
#' @param rpus vector of rpu codes to be merged
#' @param rpu_vpu_out outputs from rpus to other rpus/vpus
#' @param ref_gpkg_fun function to convert a rpu into a refactored gpkg path
#' @param agg_gpkg_fun function to convert a rpu into an aggregate gpkg path
#' @param lookup_table_refactor table name to get the lookup table from
#' @param reconciled_layer table name to get the reconciled flowlines from
#' @param divide_layer table name to get the reconciled divides from
#' @param split_divide_layer table name to get the split catchment divides from
#' @param agg_fline_layer table name to get the aggregate flowlines from
#' @param agg_cats_layer table name to get the aggregate catchments from
#' @param mapped_oulets_layer table name to get the long form mapped pois from
#' 
merge_refactor <- function(rpus, 
                           rpu_vpu_out, 
                           ref_gpkg_fun,
                           agg_gpkg_fun,
                           lookup_table_refactor, 
                           reconciled_layer,
                           divide_layer, 
                           split_divide_layer,
                           agg_fline_layer,
                           agg_cats_layer,
                           mapped_outlets_layer) {
  
  
  out <- rep(list(list()), length(rpus))
  names(out) <- rpus
  
  s <- 10000000
  
  for(rpu in rpus) {
    print(rpu)
    
    refactor_gpkg <- ref_gpkg_fun(rpu)
    aggregate_gpkg <- agg_gpkg_fun(rpu)
    
    out[[rpu]] <- setNames(list(read_sf(aggregate_gpkg, lookup_table_refactor), 
                                read_sf(refactor_gpkg, reconciled_layer), 
                                read_sf(refactor_gpkg, divide_layer), 
                                read_sf(refactor_gpkg, split_divide_layer),
                                read_sf(aggregate_gpkg, agg_fline_layer), 
                                read_sf(aggregate_gpkg, agg_cats_layer), 
                                read_sf(aggregate_gpkg, mapped_outlets_layer)),
                           c(lookup_table_refactor, 
                             reconciled_layer, 
                             divide_layer, 
                             split_divide_layer,
                             agg_fline_layer, 
                             agg_cats_layer, 
                             mapped_outlets_layer))#,
    
    ### RECONCILED ###
    
    out[[rpu]][[reconciled_layer]] <- select(out[[rpu]][[reconciled_layer]], 
                                             ID, toID, LENGTHKM, TotDASqKM, member_COMID, 
                                             LevelPathID = orig_levelpathID)
    
    out[[rpu]][[reconciled_layer]]$newID <- 
      seq(s, by = 1, length.out = nrow(out[[rpu]][[reconciled_layer]]))
    
    s <- max(out[[rpu]][[reconciled_layer]]$newID) + 1
    
    # get a toID that matches the new IDs
    out[[rpu]][[reconciled_layer]] <- left_join(out[[rpu]][[reconciled_layer]], 
                                                select(st_drop_geometry(out[[rpu]][[reconciled_layer]]), 
                                                       ID, newtoID = newID), 
                                                by = c("toID" = "ID"))
    
    # find updates that we need to apply
    update <- rpu_vpu_out[!rpu_vpu_out$toCOMID %in% out[[rpu]][[lookup_table_refactor]]$NHDPlusV2_COMID &
                            rpu_vpu_out$COMID %in% out[[rpu]][[lookup_table_refactor]]$NHDPlusV2_COMID, ]
    
    # apply these updates for follow up if we got any
    if(nrow(update) > 0) {
      update <- left_join(update, out[[rpu]][[lookup_table_refactor]], 
                          by = c("COMID" = "NHDPlusV2_COMID")) %>%
        left_join(select(out[[rpu]][[lookup_table_refactor]],
                         NHDPlusV2_COMID, to_member_COMID = member_COMID), 
                  by = c("toCOMID" = "NHDPlusV2_COMID"))
      
      out[[rpu]][[reconciled_layer]] <- left_join(out[[rpu]][[reconciled_layer]], 
                                                  select(update, reconciled_ID, toCOMID),
                                                  by = c("ID" = "reconciled_ID"))
    }
    
    #### LOOKUP TABLE ####
    
    update_id <- st_drop_geometry(select(out[[rpu]][[reconciled_layer]], 
                                         old_ID = ID, ID = newID))
    
    out[[rpu]][[lookup_table_refactor]] <- out[[rpu]][[lookup_table_refactor]] %>%
      left_join(update_id, by = c("reconciled_ID" = "old_ID")) %>%
      select(-reconciled_ID, reconciled_ID = ID) %>%
      left_join(update_id, by = c("aggregated_flowpath_ID" = "old_ID")) %>%
      select(-aggregated_flowpath_ID, aggregated_flowpath_ID = ID) %>%
      left_join(update_id, by = c("aggregated_divide_ID" = "old_ID")) %>%
      select(-aggregated_divide_ID, aggregated_divide_ID = ID) %>%
      left_join(st_drop_geometry(select(out[[rpu]][[reconciled_layer]], 
                                        ID = newID, LevelPathID)),
                by = c("reconciled_ID" = "ID"))
    
    #### DIVIDE LAYER ####
    
    out[[rpu]][[divide_layer]] <- left_join(rename(out[[rpu]][[divide_layer]], 
                                                   old_ID = ID),
                                            update_id, by = c("old_ID"))
    
    out[[rpu]][[divide_layer]] <- select(out[[rpu]][[divide_layer]], 
                                         ID, areasqkm, member_COMID)
    
    out[[rpu]][[divide_layer]]$rpu <- rep(rpu, nrow(out[[rpu]][[divide_layer]]))
    
    #### aggregate layers ####
    
    for(l in c(agg_cats_layer, agg_fline_layer)) { #, mapped_outlets_layer
      
      out[[rpu]][[l]] <- left_join(rename(out[[rpu]][[l]], old_ID = ID),
                                   update_id, 
                                   by = c("old_ID"))
      
      if("toID" %in% names(out[[rpu]][[l]])) {
        out[[rpu]][[l]] <- left_join(rename(out[[rpu]][[l]], old_toID = toID),
                                     rename(update_id, toID = ID), 
                                     by = c("old_toID" = "old_ID"))
        
        sets <- select(st_drop_geometry(out[[rpu]][[l]]), ID, old_ID, old_set = set) %>%
          mutate(old_set = strsplit(old_set, ",")) %>%
          tidyr::unnest(cols = old_set) %>%
          mutate(old_set = as.integer(old_set)) %>%
          left_join(rename(update_id, set = ID), by  = c("old_set" = "old_ID")) %>%
          select(-old_set, -old_ID)
        
        sets <- split(sets$set, sets$ID)
        
        sets <- data.frame(ID = as.integer(names(sets))) %>%
          mutate(set = unname(sets))
        
        out[[rpu]][[l]] <- left_join(select(out[[rpu]][[l]], -set),
                                     sets, by = "ID")
        
        out[[rpu]][[l]] <- select(out[[rpu]][[l]], -old_ID, -old_toID)
        
        out[[rpu]][[l]] <- out[[rpu]][[l]][c("ID", "toID", "set", 
                                             names(out[[rpu]][[l]])[!names(out[[rpu]][[l]]) %in% 
                                                                      c("ID", "toID", "set", 
                                                                        attr(out[[rpu]][[l]], "sf_column"))],
                                             attr(out[[rpu]][[l]], "sf_column"))]
        
        out[[rpu]][[l]]$set <- sapply(out[[rpu]][[l]]$set, paste, collapse = ",")
      } else {
        
        out[[rpu]][[l]] <- select(out[[rpu]][[l]], -old_ID)
        
        out[[rpu]][[l]] <- out[[rpu]][[l]][c("ID", 
                                             names(out[[rpu]][[l]])[!names(out[[rpu]][[l]]) %in% 
                                                                      c("ID", 
                                                                        attr(out[[rpu]][[l]], "sf_column"))],
                                             attr(out[[rpu]][[l]], "sf_column"))]
      }
    }
    
  }
  
  out <- setNames(lapply(names(out[[1]]), function(x, out) {
    dplyr::bind_rows(lapply(out, function(df, n) {
      if("COMID" %in% names(df[[n]]) && is.numeric(df[[n]]$COMID)) 
        df[[n]]$COMID <- as.character(df[[n]]$COMID)
      df[[n]]
      
      if("event_id" %in% names(df[[n]]) && !is.character(df[[n]]$event_id))
        df[[n]]$event_id <- as.character(df[[n]]$event_id)
      df[[n]]
    }, n = x))
  }, out = out), names(out[[1]]))
  
  if("toCOMID" %in% names(out[[reconciled_layer]])) {
    
    # blow up so we have unique COMIDs to join on.
    # need to keep the top most of any splits (the .1 variety)
    # this makes sure out toCOMID assignments go to the right new id.
    long_form <- st_drop_geometry(out[[reconciled_layer]]) %>%
      select(newID, member_COMID) %>%
      mutate(member_COMID = strsplit(member_COMID, ",")) %>%
      tidyr::unnest(cols = member_COMID) %>%
      filter(grepl("\\.1$", member_COMID) | !grepl("\\.", member_COMID)) %>%
      mutate(NHDPlusV2_COMID = as.integer(member_COMID)) %>%
      select(-member_COMID, update_newtoID = newID)
    
    # NOTE: if the missing tocomid is in the next VPU they will still be NA
    out[[reconciled_layer]] <- left_join(out[[reconciled_layer]], 
                                         long_form, by = c("toCOMID" = "NHDPlusV2_COMID"))
    out[[reconciled_layer]]$newtoID[!is.na(out[[reconciled_layer]]$toCOMID)] <- 
      out[[reconciled_layer]]$update_newtoID[!is.na(out[[reconciled_layer]]$toCOMID)]
    
    # now use updates from refactored ids with aggregate
    long_form <- st_drop_geometry(out[[agg_fline_layer]]) %>%
      select(ID, set) %>%
      mutate(set = strsplit(set, ",")) %>%
      tidyr::unnest(cols = set) %>%
      mutate(set = as.integer(set)) %>%
      select(update_newtoID = ID, refactor_ID = set)
    
    # join updated toID from refactor so we can update as needed
    out[[agg_fline_layer]] <- out[[agg_fline_layer]] %>%
      # this adds a "new_toID" containing the refactor id that the 
      # aggregate flowpaths we are updating need to go to.
      left_join(select(st_drop_geometry(out[[reconciled_layer]]), 
                       ID = newID, new_toID = update_newtoID),
                by = "ID") %>%
      # we need to handle sets of refactor ids and get the right aggregate toID
      left_join(long_form, by = c("new_toID" = "refactor_ID"))
    # need to update the field to reflect the information joined just above.
    out[[agg_fline_layer]]$new_toID[!is.na(out[[agg_fline_layer]]$update_newtoID)] <-
      out[[agg_fline_layer]]$update_newtoID[!is.na(out[[agg_fline_layer]]$update_newtoID)]
    
    # now do the actual toID updates
    out[[agg_fline_layer]]$toID[!is.na(out[[agg_fline_layer]]$new_toID)] <- 
      out[[agg_fline_layer]]$new_toID[!is.na(out[[agg_fline_layer]]$new_toID)]
    
    out[[agg_fline_layer]] <- select(out[[agg_fline_layer]], -new_toID, -update_newtoID)
  }
  
  out[[split_divide_layer]] <- rename(out[[split_divide_layer]], comid_part = FEATUREID)
  
  out[[reconciled_layer]] <- select(out[[reconciled_layer]], ID = newID, 
                                    toID = newtoID, LENGTHKM, TotDASqKM, 
                                    member_COMID, LevelPathID, refactor_ID = ID)
  
  out
}

#' get vpu configuration
#' @param vpu vpu to get configuration for
#' @param data_paths list of paths to data resources
#' 
get_vpu_config <- function(vpu, data_paths, full_nhd, full_wbd) {
  
  elev <- data_paths$elev_cm[grepl(paste0("Ned", substr(vpu, 1, 2)), 
                                   data_paths$elev_cm, ignore.case = TRUE)]
  
  if (vpu == "08"){
    elev$rpu_03g <- data_paths$elev_cm$rpu_03g
  }
  
  if(vpu == "02"){
    grep_exp <-"^02|^04"
  } else if (vpu == "08") {
    grep_exp <- "^03|^08"
  } else {
    grep_exp <- paste0("^", substr(vpu, start = 1, stop = 2))
    elev <- append(elev, list(rpu_03g = data_paths$elev_cm$rpu_03g))
  }
  
  vpu_nhd <- full_nhd %>%
    filter(grepl(paste0("^", grep_exp, ".*"), .data$VPUID)) %>%
    nhdplusTools::align_nhdplus_names(.)
  
  vpu_WBD <- full_wbd %>%
    filter(grepl(paste0("^", grep_exp, ".*"), .data$HUC_12))
  
  list(vpu_nhd = vpu_nhd, vpu_WBD = vpu_WBD, elev = elev)
}

#' make hu12 xwalk
#' @param cats catchments from reference fabric
#' @param divides reconciled divides from refactor
#' @param vpu vpu that we are running
#' @param wbd_xwalk crosswalk input file
#' @param wbd_sf raw wbd dataset
make_hu12_xwalk <- function(cats, divides, vpu, wbd_xwalk, wbd_sf) {
  
  # Read in full NHD cats
  cats <- filter(cats, full_cats == 1)
  cats <- sf::st_make_valid(cats)
  divides <- sf::st_make_valid(divides)
  
  # Intersect NHD catchments and divides with HUC12
  nhd_wbd_int <- get_HUC12_Xwalk(vpu, cats, divides,
                                 wbd_xwalk,
                                 wbd_sf)
  
  # Bring over divides/HUC12 intersection information into divides layer
  xwalk_divides_wbd <- st_drop_geometry(nhd_wbd_int$divides_HUC12) %>%
    select(-c(ACRES, HU_12_MOD))
  
  divides <- divides %>%
    left_join(distinct(xwalk_divides_wbd) %>%
                group_by(ID) %>%
                filter(intArea == max(intArea)) %>%
                ungroup() %>%
                select(ID, HUC_12_int, intArea, AreaHUC12), 
              by = "ID") 
  
  # Bring over divides/HUC12 intersection information into divides layer
  xwalk_nhd_wbd <- st_drop_geometry(nhd_wbd_int$cats_HUC12) %>%
    select(-c(ACRES, HU_12_MOD))
  
  rm(nhd_wbd_int)
  
  cats <- cats %>%
    left_join(distinct(xwalk_nhd_wbd) %>%
                group_by(FEATUREID) %>%
                slice(which.max(intArea)) %>%
                ungroup() %>%
                select(FEATUREID, HUC_12_int, intArea, AreaHUC12),
              by = "FEATUREID") %>%
    st_transform(proj_crs) 
  
  # # All intersecting NHD cat / divides / HUC12 geometries (1: many)
  # write_sf(xwalk_nhd_wbd, ND_gpkg, xwalk_layer)
  # write_sf(xwalk_divides_wbd, ND_gpkg, divides_xwalk)
  # 
  # # Cats with the HUC_12 value (HUC_12_int) that is the max overlap
  # write_sf(cats, ND_gpkg, nhd_catchment)
  # write_sf(divides, ND_gpkg, divide_layer)
  
  list(xwalk_layer = xwalk_nhd_wbd, divides_xwalk = xwalk_divides_wbd, 
       nhd_catchment = cats, divide_layer = divides)
  
}

#' Merges geopackages together to create CONUs geopackage of features
#' @param VPU VPU from NHDPlusV2
#' @param full_cats  (sf data.frame) all catchments (sinks/flowline) within given VPU
#' @param divides (sf data.frame) divides layer for a given VPU
#' @param wbd  (data.frame) HUC12 X-walk table, see get_data for full citation
#' @param wbd_SF (sf data.frame) HUC12 layer from NHDPlusv2.1 national geodatabase
#' @return (sf data.frame) intersection of nhd catchments and divides layer with HUC12
#'
get_HUC12_Xwalk <- function(vpu, full_cats = FALSE, divides = FALSE, wbd = FALSE, wbd_SF = FALSE){
  
  # HUC02 includes some 
  if(vpu == "02"){
    grep_exp <-"^02|^04"
  } else if (vpu == "08") {
    grep_exp <- "^03|^08"
  } else {
    grep_exp <- paste0("^", substr(vpu, start = 1, stop = 2))
  }
  
  # crosswalk to HU12, filter by VPU
  nhd_to_HU12 <- read.csv(wbd, colClasses = c("character", "integer", "character")) %>% 
    filter(grepl(grep_exp, .data$HUC_12)) %>%
    mutate(HUC_8 = substr(HUC_12, 1, 8))
  
  # Read in the NHDPlus National GDB HUC12 layer
  HUC12_lyr <- wbd_SF %>% 
    #filter(grepl(paste0("^", VPU,".*"), .data$HUC_12)) %>%
    filter(grepl(grep_exp, .data$HUC_12)) %>% #, 
    st_transform(st_crs(full_cats)) %>%
    st_make_valid() %>%
    select(HUC_12, ACRES, HU_10_TYPE, HU_12_DS, HU_12_MOD, HU_12_TYPE)
  
  reg_cats <- full_cats %>% 
    left_join(select(nhd_to_HU12, c(FEATUREID, HUC_12)), by = "FEATUREID")
  
  sqm_per_sqkm <- 0.000001
  acres_per_sqkm <- .00404686
  # Intersect the HUC12s and catchments
  cats_HUC12 <- sf::st_intersection(reg_cats, HUC12_lyr) %>%
    # Convert area to squaker kilometers 
    mutate(intArea = as.numeric(st_area(.)) * sqm_per_sqkm,
           AreaHUC12 = ACRES * acres_per_sqkm) %>%
    rename(HUC_12_int = HUC_12.1)
  
  divides_HUC12 <- sf::st_intersection(divides, HUC12_lyr) %>%
    # Convert area to squaker kilometers 
    mutate(intArea = as.numeric(st_area(.)) * sqm_per_sqkm,
           AreaHUC12 = ACRES * acres_per_sqkm) %>%
    rename(HUC_12_int = HUC_12)
  
  
  return(list(cats_HUC12 = cats_HUC12, divides_HUC12 = divides_HUC12))
}

#' make divides nd
#' @param divides divides after hu12 crosswalk
#' @param lookup lookup_table_aggregate from merged vpus
#' @param cats nhd_catchment from hu12 crosswalk step
#' @param cat_rpu_table table listing what rpu each catchment belongs to
#' 
make_divides_nd <- function(divides, lookup, cats, cat_rpu_table, debug = FALSE) {
  
  # Bind divide with lookup, identify aggregation step
  divides_lu <- divides %>%
    left_join(distinct(select(lookup, reconciled_ID, aggregated_divide_ID)), 
              by = c("ID" = "reconciled_ID")) %>%
    filter(!is.na(ID) & !is.na(aggregated_divide_ID)) %>%
    rename(POI_ID = aggregated_divide_ID) %>%
    # attribute that hyrefactor was the step used to aggregate these catchments
    mutate(aggStep = "hyRef") 
  
  # Identify cats witin the full catchment set not refactored or aggregated, add
  #          to divides data frame
  cats_miss <- cats %>%
    left_join(select(lookup, NHDPlusV2_COMID, member_COMID, reconciled_ID , 
                     POI_ID = aggregated_divide_ID),
              by = c("FEATUREID" =  "NHDPlusV2_COMID")) %>%
    filter(is.na(POI_ID), !member_COMID %in% divides_lu$member_COMID) %>%
    mutate(aggStep = NA) %>%
    distinct() %>%
    left_join(select(cat_rpu_table, FEATUREID, rpu = RPUID)) %>%
    #filter(!reconciled_ID %in% divides_lu$ID) %>%
    select(-member_COMID) %>%
    mutate(areasqkm = as.numeric(st_area(.)/1000000)) %>%
    select(ID = reconciled_ID, member_COMID = FEATUREID, rpu, HUC_12_int, intArea, AreaHUC12, POI_ID, aggStep)
  
  divides_lu <- divides_lu %>%
    select(-areasqkm) %>%
    rbind(cats_miss) %>%
    # resolve terminal non-POI outlets where we can
    nhdplusTools:::check_valid() %>%
    st_cast("POLYGON")
  
  # clean_geometry(divides_lu, ID = "POI_ID", keep = 1.0, crs = st_crs(divides_lu))
  divides_lu <- dissolve_holes(divides_lu)
  
  rm(cats_miss)
  
  # DEBUG:
  # HRU layer
  protoHRUs <- NULL
  if(debug) {
    protoHRUs <- divides_lu %>%
      group_by("POI_ID") %>%
      summarize(do_union = TRUE) %>%
      sfheaders::sf_remove_holes(.) %>%
      st_make_valid()
    # write_sf(protoHRUs, ND_gpkg, HRU_layer)
    # rm(protoHRUs) 
  }
  
  # write_sf(divides_lu, ND_gpkg, divides_nd)
  
  list(divides_nd = divides_lu, HRU_layer = protoHRUs)
}

#' Dissolves internal unaggregated divides within existing aggregated catchments
#' @param divides_poi (sf data frame) data frame of reconciled divides
#
#' @return (sf data frame) data frame with internal unaggregated divides assigned
#'                         to surrounding aggregated catchment
dissolve_holes <- function(divides_poi){
  # Catchments without a POI_ID or aggregated ID assigned
  miss_cats <- filter(divides_poi, is.na(POI_ID))
  
  # HRU layer
  protoHRUs <- filter(divides_poi, !is.na(POI_ID)) %>%
    group_by(POI_ID) %>%
    summarize(do_union = TRUE) %>%
    clean_geometry(., ID = "POI_ID")
  
  agg_noholes <- sfheaders::sf_remove_holes(protoHRUs) %>%
    st_cast("MULTIPOLYGON")
  
  # centroids of missing HRUs
  cents <- st_point_on_surface(miss_cats)
  # intersect centroids and HRUs
  cats_int <- st_intersects(cents, agg_noholes)
  
  # Index of HRUs that missing cats may lie within with holes filled
  POI_ID_feat <- sapply(cats_int, function(s) if (length(s) == 0) NA else s)
  
  # Return if no results
  if(sum(is.na(POI_ID_feat)) == length(POI_ID_feat)){
    return(divides_poi)
  }
  
  # Populate ID of HRUs where applicable
  miss_cats$POI_ID <- protoHRUs[POI_ID_feat,]$POI_ID 
  
  miss_cats_tobind <- filter(miss_cats, !is.na(POI_ID)) %>%
    mutate(aggStep = "internal")
  
  # bring over IDs to full data frame
  divides_poi <- divides_poi %>%
    left_join(select(st_drop_geometry(miss_cats_tobind), 
                     member_COMID, new_POI_ID = POI_ID, new_aggStep = aggStep), 
              by = "member_COMID") %>%
    mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
           aggStep = ifelse(!is.na(new_POI_ID), new_aggStep, aggStep)) %>%
    select(-c(new_POI_ID, new_aggStep)) 
  
  return(divides_poi)
}

#' assign huc12 function
#' @param divides_lu divides from divides non dritic step
#' @param xwalk_nhd_wbd xwalk_layer from huc12 crosswalk step
#' @param vpu_nhd vpu_nhd from vpu_attributes step
#' @param debug logical whether to output proto HRUs
#' @param CAC_thresh coefficient of areal coorespondence -- minimum measure of common 
#' footprint between two features taking into acount
#' the area of each feature in proporation to overlap area
assign_huc12_fun <- function(divides_lu, xwalk_nhd_wbd, vpu_nhd, debug = FALSE, CAC_thresh = 0.66) {
  
  print(paste0("Currently there are ", sum(is.na(divides_lu$POI_ID)), " cats without a POI_ID"))
  
  # This will crate aggregated IDs for NHD catchments that match HUC12 footprint if aggregated together
  # The Coefficient of Areal correpsondence - is the last argument
  if("AREASQKM" %in% names(xwalk_nhd_wbd)) {
    xwalk_nhd_wbd <- rename(xwalk_nhd_wbd,
                            AreaSqKM = AREASQKM)
  }
  
  divides_lu <- assign_HUC12(divides_lu, xwalk_nhd_wbd, 
                             filter(vpu_nhd, FTYPE != "Coastline"), CAC_thresh)
  
  # Determine if any unaggregated catchments match HUC10 footprint (mostly in the west)
  divides_lu <- assign_HUC10(divides_lu, xwalk_nhd_wbd,
                             filter(vpu_nhd, FTYPE != "Coastline"), CAC_thresh) %>%
    select(-comid_char)
  
  protoHRUs <- NULL
  if(debug) {
    # DEBUG:
    # HRU layer
    # this is not working in my testing
    protoHRUs <- divides_lu %>%
      group_by(POI_ID) %>%
      summarize(do_union = TRUE) %>%
      sfheaders::sf_remove_holes(.) %>%
      st_make_valid()
    # write_sf(protoHRUs, ND_gpkg, HRU_layer)
    # rm(protoHRUs)
  }
  
  # Update missing_cats
  # write_sf(divides_lu, ND_gpkg, divides_nd)
  
  list(divides_nd = divides_lu, HRU_layer = protoHRUs)
}

#' Determine if any divides that have not been assigned to an aggregated catchment 
#'           can be aggregated on the basis of their HUC12 ID
#' @param divides (sf data frame) data frame of reconciled divides
#' @param HUC12_xwalk (data frame) data frame of reconciled divides - HUC12 intersection
#' @param nhd (sf data frame) data frame of nhd flowlines
#' @param CAC_num (numeric) coefficient of aereal correspondence; minimum measure of common footprint between two features taking into acount
#          the area of each feature in proporation to overlap area
#
#   see: Minnick, R. F. (1964). “A Method for the Measurement of Areal Correspondence.’’ Papers of the
#        Michigan Academy of Science, Arts, and Letters, XLIX, 333-43. 1963 Meeting.
#
#' @return (sf data frame) data frame with unaggregated divides assigned if HUC10 if grouped divides
#'                         meet CAC criteria
assign_HUC12 <- function(divides_poi, HUC12_xwalk, nhd, CAC_num){
  
  miss_cats_sub <- filter(divides_poi, is.na(POI_ID))
  
  # HUC12/cat intersections associated with missing catchments
  nhd_wbd_int_sub <- filter(HUC12_xwalk, FEATUREID %in% miss_cats_sub$member_COMID)
  
  nhd <- mutate(nhd, COMID = as.character(COMID))
  
  #***************************************
  # STEP 1 - Identify aggregated cats without aggregated ID that match the HUC12 footprint
  # Note - out west we may need to expand this to HUC10
  
  # HUC12s that match Catchment footprint
  huc12_in_nhd <- nhd_wbd_int_sub %>%
    # int_HUC_12 is the intersection HUC12, not the NAWQA Xwalk HUC12
    group_by(HUC_12_int) %>%
    # Which HUC12 has the largest intersecting area with a given catchment
    slice(which.max(intArea)) %>%
    ungroup() %>%
    group_by(FEATUREID, HUC_12_int) %>%
    summarize(n_NHD = n(),
              NHD_area = mean(AreaSqKM),
              HUC_12_area = sum(AreaHUC12),
              intarea = sum(intArea),
              CAC = intarea / ((NHD_area - intarea) + (HUC_12_area - intarea) + intarea)) %>%
    # Keep only catchments with a decent matching footprint with HUC12
    # only return those for which there is one HUC12 per FEATUREID
    # CAC - coefficient of area correspondence, we'll use a 0.66 measure for overlap
    filter(CAC > CAC_num, n() == 1) %>%
    ungroup()
  
  # HUC12 IDs of the HUC12s above
  HUC12_in_cats <- filter(st_drop_geometry(miss_cats_sub), HUC_12_int %in% huc12_in_nhd$HUC_12_int) %>% # was member_COMID/FEATUREID
    select(member_COMID, HUC_12_int) %>%
    rename(new_POI_ID = HUC_12_int)
  
  # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
  divides_poi <- divides_poi %>%
    left_join(select(HUC12_in_cats, member_COMID, new_POI_ID), by = "member_COMID") %>%
    mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
           aggStep = ifelse(member_COMID %in% huc12_in_nhd$FEATUREID, "HUC_12", aggStep)) %>%
    select(-new_POI_ID)
  
  #***************************************
  # STEP 2 - Identify HUC12s that match very large nhd catchments
  # Note - out west we may need to expand this to HUC10
  
  # Update the number of missing catchments
  miss_cats_sub <- filter(divides_poi, is.na(POI_ID))
  
  # NHD catchments aggregated to HUC12 that match the HUC12 footprint
  nhd_in_huc12 <- filter(nhd_wbd_int_sub, FEATUREID %in% miss_cats_sub$member_COMID) %>%
    group_by(FEATUREID) %>%
    filter(intArea == max(intArea)) %>%
    ungroup() %>%
    group_by(HUC_12_int) %>%
    summarize(n_HUC12 = n(),
              NHD_area = sum(AreaSqKM),
              HUC_12_area = mean(AreaHUC12),
              intarea = sum(intArea),
              CAC = intarea / ((NHD_area - intarea) + (HUC_12_area - intarea) + intarea)) %>%
    filter(CAC > CAC_num, !HUC_12_int %in% HUC12_in_cats$new_POI_ID)
  
  # HUC12 IDs of the HUC12s above
  cat_in_HUC12s <- filter(st_drop_geometry(miss_cats_sub), HUC_12_int %in% nhd_in_huc12$HUC_12_int) %>% 
    select(member_COMID, HUC_12_int) %>%
    rename(new_POI_ID = HUC_12_int)
  
  # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
  divides_poi <- divides_poi %>%
    left_join(select(cat_in_HUC12s, member_COMID, new_POI_ID), by = "member_COMID") %>%
    mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
           aggStep = ifelse(member_COMID %in% cat_in_HUC12s$member_COMID, "HUC_12", aggStep)) %>%
    select(-new_POI_ID)
  
  #***************************************
  # STEP 2 - Identify edge non-aggregated catchments that may contribute to the HUC12 but are not 
  #          assigned an aggregated ID and may not share the HUC12 value
  # Note - out west we may need to expand this to HUC10
  
  # Update the number of missing catchments
  miss_cats_sub <- filter(divides_poi, is.na(POI_ID))
  
  # Flowlines within the POI assigned HUC12
  inFL <- nhd %>%
    inner_join(st_drop_geometry(filter(divides_poi, aggStep == "HUC_12")) %>% 
                 select(member_COMID, POI_ID), by = c("COMID" = "member_COMID")) %>%
    select(COMID, Hydroseq, POI_ID)
  
  # Flowlines contributing to the POI assigned HUC12
  inFL_into <- nhd %>% 
    inner_join(st_drop_geometry(inFL) %>% select(-COMID), by = c("DnHydroseq" = "Hydroseq")) %>%
    select(COMID, DnHydroseq, DS_POI_ID = POI_ID) %>%
    inner_join(st_drop_geometry(divides_poi) %>% select(member_COMID, POI_ID), by = c("COMID" = "member_COMID")) %>%
    filter(is.na(POI_ID)) 
  
  # This may be 0 in some regions
  if(nrow(inFL_into) > 0){
    
    # Get upstream flowlines contributing to the HUC12 that are not assigned an aggregated catchment ID
    cats_us <- do.call("rbind", lapply(inFL_into$COMID, function(x) {
      print(x)
      seg <- nhdplusTools::get_UT(nhd, x)
      cats_to_att <- filter(miss_cats_sub, member_COMID %in% seg) %>%
        mutate(TermCom = x)
    })) 
    
    # Bring the whole things together
    cats_us_POI_ID <- select(st_drop_geometry(cats_us), member_COMID, TermCom) %>%
      inner_join(st_drop_geometry(inFL_into), by = c("TermCom" = "COMID")) %>%
      rename(new_POI_ID = DS_POI_ID) %>%
      select(member_COMID, new_POI_ID)
    
    # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
    divides_poi <- divides_poi %>%
      left_join(cats_us_POI_ID, by = "member_COMID") %>%
      mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
             aggStep = ifelse(member_COMID %in% cats_us_POI_ID$member_COMID, "HUC_12", aggStep)) %>%
      select(-new_POI_ID)
    
  }
  return(divides_poi)
  
}

#' Determine if any divides that have not been assigned to an aggregated catchment 
#'           can be aggregated on the basis of their HUC10 ID
#' @param divides (sf data frame) data frame of reconciled divides
#' @param HUC12_xwalk (data frame) data frame of reconciled divides - HUC12 intersection
#' @param nhd (sf data frame) data frame of nhd flowlines
#' @param CAC_num (numeric) coefficient of aereal correspondence; minimum measure of common footprint between two features taking into acount
#          the area of each feature in proporation to overlap area
#
#   see: Minnick, R. F. (1964). “A Method for the Measurement of Areal Correspondence.’’ Papers of the
#        Michigan Academy of Science, Arts, and Letters, XLIX, 333-43. 1963 Meeting.
#
#' @return (sf data frame) data frame with unaggregated divides assigned if HUC10 if grouped divides
#'                         meet CAC criteria
assign_HUC10 <- function(divides, HUC12_xwalk, nhd, CAC_num){
  miss_cats_sub <- filter(divides, is.na(POI_ID)) %>%
    mutate(HUC_10 = substr(HUC_12_int, 1, 10))
  
  nhd <- mutate(nhd, COMID = as.character(COMID))
  
  nhd_wbd_int_sub <- HUC12_xwalk %>%
    mutate(HUC_10 = substr(HUC_12_int, 1, 10)) 
  
  HUC10_area <- nhd_wbd_int_sub %>%
    distinct(HUC_10, AreaHUC12, .keep_all = T) %>%
    group_by(HUC_10) %>%
    summarize(AreaHUC10 = sum(AreaHUC12)) %>%
    ungroup()
  
  nhd_wbd_int_sub <- nhd_wbd_int_sub %>%
    left_join(HUC10_area, by = "HUC_10")
  
  # HUC12/cat intersections associated with missing catchments
  nhd_wbd_int_sub <- filter(nhd_wbd_int_sub, FEATUREID %in% miss_cats_sub$member_COMID)
  
  #***************************************
  # STEP 1 - Identify nhd cats that match huc10 footrpint
  # Note - out west we may need to expand this to HUC10
  
  # HUC12s that match Catchment footprint
  huc10_in_nhd <- nhd_wbd_int_sub %>%
    # HUC_12_int is the intersection HUC12
    group_by(HUC_12_int) %>%
    # Which HUC12 has the largest intersecting area with a given catchment
    slice(which.max(intArea)) %>%
    ungroup() %>%
    group_by(HUC_10) %>%
    mutate(n_NHD = n(),
           NHD_area = mean(AreaSqKM),
           HUC_10_area = mean(AreaHUC10),
           intarea = sum(intArea),
           CAC = intarea / ((NHD_area - intarea) + (HUC_10_area - intarea) + intarea)) %>%
    # Keep only catchments with a decent matching footprint with HUC12
    # CAC - coefficient of area correspondence, we'll use a 0.66 measure for overlap
    filter(CAC > CAC_num) 
  
  # HUC12 IDs of the HUC12s above
  HUC10_in_cats <- filter(st_drop_geometry(miss_cats_sub), member_COMID %in% huc10_in_nhd$FEATUREID) %>%
    select(member_COMID, HUC_10) %>%
    rename(new_POI_ID = HUC_10)
  
  # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
  divides <- divides %>%
    left_join(select(HUC10_in_cats, member_COMID, new_POI_ID), by = "member_COMID") %>%
    mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
           aggStep = ifelse(member_COMID %in% huc10_in_nhd$FEATUREID, "HUC_10", aggStep)) %>%
    select(-new_POI_ID)
  
  #***************************************
  # STEP 2 - Identify edge non-aggregated catchments that may contribute to the HUC12 but are not 
  #          assigned an aggregated ID and may not share the HUC12 value
  # Note - out west we may need to expand this to HUC10
  
  # Update the number of missing catchments
  miss_cats_sub <- filter(divides, is.na(POI_ID))
  
  # Flowlines within the POI assigned HUC12
  inFL <- nhd %>%
    inner_join(st_drop_geometry(filter(divides, aggStep == "HUC_10")) %>% 
                 select(member_COMID, POI_ID), by = c("COMID" = "member_COMID")) %>%
    select(COMID, Hydroseq, POI_ID)
  
  # Flowlines contributing to the POI assigned HUC12
  inFL_into <- nhd %>% 
    inner_join(st_drop_geometry(inFL) %>% select(-COMID), by = c("DnHydroseq" = "Hydroseq")) %>%
    select(COMID, DnHydroseq, DS_POI_ID = POI_ID) %>%
    inner_join(st_drop_geometry(divides) %>% select(member_COMID, POI_ID), by = c("COMID" = "member_COMID")) %>%
    filter(is.na(POI_ID)) 
  
  # This may be 0 in some regions
  if(nrow(inFL_into) > 0){
    
    # Get upstream flowlines contributing to the HUC12 that are not assigned an aggregated catchment ID
    cats_us <- do.call("rbind", lapply(inFL_into$COMID, function(x) {
      print(x)
      seg <- nhdplusTools::get_UT(nhd, x)
      cats_to_att <- filter(miss_cats_sub, member_COMID %in% seg) %>%
        mutate(TermCom = x)
    })) 
    
    # Bring the whole things together
    cats_us_POI_ID <- select(st_drop_geometry(cats_us), member_COMID, TermCom) %>%
      inner_join(st_drop_geometry(inFL_into), by = c("TermCom" = "COMID")) %>%
      rename(new_POI_ID = DS_POI_ID) %>%
      select(member_COMID, new_POI_ID)
    
    # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
    divides <- divides %>%
      left_join(cats_us_POI_ID, by = "member_COMID") %>%
      mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID),
             aggStep = ifelse(member_COMID %in% cats_us_POI_ID$member_COMID, "HUC_10", aggStep)) %>%
      select(-new_POI_ID)
    
  }
  
  divides <- dissolve_holes(divides)
  
  return(mutate(divides, comid_char = as.character(member_COMID)))
}

#' create coastal catchments
#' @param vpu_nhd vpu_nhd from vpu_attributes step
#' @param divides_lu divides from assign huc12 step
#' @param vpu_WBD vpu_WBD from vpu_attributes step
#' @param debug logical whether to output proto HRUs
#' 
create_coastal_catchments <- function(vpu_nhd, divides_lu, vpu_WBD, debug = FALSE) {
  protoHRUs <- NULL
  # aggregate frontal hucs of same type
  if("Coastline" %in% unique(vpu_nhd$FTYPE)){
    print(paste0("Currently there are ", sum(is.na(divides_lu$POI_ID)), " cats without a POI_ID"))
    
    
    # Function to create coastal catchments by HUC12 and catch inflows.
    divides_lu <- coastal_cats(divides_lu, vpu_nhd, vpu_WBD)
    
    if(debug) {
      # DEBUG:
      # HRU layer
      protoHRUs <- divides_lu %>%
        group_by(POI_ID) %>%
        summarize(do_union = TRUE) %>%
        sfheaders::sf_remove_holes(.) %>%
        st_make_valid()
    }
    
    
    # Update missing_cats
    # write_sf(divides_lu, ND_gpkg, divides_nd)
    
  } else {
    print ("No Coastlines in VPU")
  }
  
  vpu_nhd <- filter(vpu_nhd, FTYPE != "Coastline")
  
  list(vpu_nhd = vpu_nhd, divides_nd = divides_lu, protoHRUs = protoHRUs)
}

#' Determine if any divides that have not been assigned to an aggregated catchment 
#'           can be aggregated as coastal catchments using their HUC12 ID (no CAC match requirement)
#' @param nhd (sf data frame) full data frame of nhd flowlines (including coastal flowlines)
#' @param divides (sf data frame) data frame of reconciled divides
#
#' @return (sf data frame) Coastal catchments aggregated using HUC12 ID
coastal_cats <- function(divides_poi, full_nhd, vpu_HUC12){
  # Subset to just the coastal flowlines
  coastFL <- filter(st_drop_geometry(full_nhd), FTYPE == "Coastline")
  
  divides_miss <- filter(divides_poi, is.na(POI_ID)) %>%
    mutate(member_COMID_int = as.integer(member_COMID))
  
  # If there are coastal catchments
  if(nrow(coastFL) == 0){
    return(divides)
  } else {
    
    # Missing cats that are coastal cats
    coastal_cats <- filter(divides_miss, member_COMID_int %in% coastFL$COMID) 
    
    # Extract unaggregated cats within frontal HUC12s that may not be assigned later
    frontal_HUC12 <- filter(vpu_HUC12, HU_12_TYPE == "F" | HU_12_DS == "OCEAN")
    
    # These next two section created non-correct aggregations in some areas, commenting out for now
    #frontal_HUC12_cats <- filter(divides_poi, HUC_12_int %in% frontal_HUC12$HUC_12, is.na(POI_ID)) %>%
    #  mutate(member_COMID = as.integer(member_COMID))
    
    #divides_miss <- rbind(coastal_cats, filter(frontal_HUC12_cats, !member_COMID %in% coastal_cats$member_COMID)) %>%
    #  mutate(member_COMID = as.character(member_COMID)) 
    
    coastalFL <- rbind(coastFL, filter(st_drop_geometry(full_nhd), COMID %in% divides_miss$member_COMID)) %>%
      distinct()
    
    # # Just assign remaining nonPOI costal cats the HUC12, not going to get much better information
    divides_miss <- divides_miss %>%
      mutate(HUC_12_POI = ifelse(member_COMID %in% coastal_cats$member_COMID, HUC_12_int, POI_ID))
    
    # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
    divides_poi <- divides_poi %>%
      left_join(st_drop_geometry(divides_miss) %>% select(member_COMID, HUC_12_POI), by = "member_COMID") %>%
      mutate(POI_ID = ifelse(!is.na(HUC_12_POI), HUC_12_POI, POI_ID)) %>%
      mutate(aggStep = ifelse(!is.na(HUC_12_POI), "coastal", aggStep)) %>%
      select(-HUC_12_POI, member_COMID, ID, rpu, POI_ID, aggStep, HUC_12_int, AreaHUC12, intArea) %>%
      mutate(vpu_ID = substr(rpu, 1, 2))
    
    # FL of cats that "contribtue" to the coast "flowlines" but are not coastal themselves
    nhd2coast <- filter(full_nhd, !COMID %in% coastFL$COMID & DnHydroseq %in% coastFL$Hydroseq) %>%
      mutate(COMID = as.character(COMID))
    
    # The coatal flowlines
    nhd_at_coast <- filter(full_nhd, Hydroseq %in% nhd2coast$DnHydroseq)
    
    # cats with flowline contributing to coast
    cat2coast <- filter(divides_miss, member_COMID %in% nhd2coast$COMID, is.na(POI_ID)) %>%
      inner_join(st_drop_geometry(nhd2coast) %>% 
                   select(COMID, DnHydroseq), by = c("member_COMID" = "COMID"))
    
    # Cats associated with the coastal flowlines
    cat_at_coast <- coastal_cats %>%
      inner_join(st_drop_geometry(nhd_at_coast) %>% 
                   select(COMID, Hydroseq), by = c("member_COMID_int" = "COMID"))
    
    nhd2coast_us <- full_nhd %>%
      filter(TerminalPa %in% filter(nhd2coast, COMID %in% cat2coast$member_COMID)$Hydroseq) %>%
      select(COMID, TerminalPa)
    
    # Assign POIs for these guys
    cat2coast_POI <- cat2coast %>%
      inner_join(st_drop_geometry(cat_at_coast) %>% select(member_COMID, Hydroseq, HUC_12_int, POI_ID), by = c("DnHydroseq" = "Hydroseq")) %>%
      # Assign POI_ID of attached coastal HUC if it exists
      mutate(POI_ID.x = ifelse(!is.na(POI_ID.y),  POI_ID.y, POI_ID.x)) %>%
      mutate(POI_ID.x = ifelse(HUC_12_int.x == HUC_12_int.y & is.na(POI_ID.x), HUC_12_int.x, POI_ID.x)) %>%
      mutate(POI_ID.x = ifelse(is.na(POI_ID.x), HUC_12_int.x, POI_ID.x))
    
    # Get upstream flowlines contributing to the coast that are not assigned a POI
    nhd2coast_POI <- filter(nhd2coast, COMID %in% cat2coast_POI$member_COMID.x)# & StartFlag != 1)
    cats_us <- do.call("rbind", lapply(nhd2coast_POI$COMID, function(x) {
      #print(x)
      seg <- nhdplusTools::get_UT(full_nhd, x)
      cats_to_att <- filter(divides_miss, member_COMID %in% seg,
                            member_COMID %in% filter(divides_miss, is.na(POI_ID))$member_COMID) %>%
        mutate(TermCom = x)
    })) 
    
    if (length(cats_us) > 0){
      # Boil down to COMID and POI_ID
      cat2coast_POI_ID <- st_drop_geometry(cat2coast_POI) %>% select(member_COMID.x, POI_ID.x) %>%
        rename(member_COMID = member_COMID.x, POI_ID = POI_ID.x)
      
      # Bring the whole things together
      cats_us_POI_ID <- select(cats_us, member_COMID, TermCom) %>% 
        inner_join(cat2coast_POI_ID, by = c("TermCom" = "member_COMID")) %>%
        rename(new_POI_ID = POI_ID) %>%
        mutate(member_COMID = as.character(member_COMID))
      
      # attribute cats POI_ID with HUC12 POI_IDs if match HUC12s
      divides_poi <- divides_poi %>%
        left_join(st_drop_geometry(cats_us_POI_ID) %>% select(member_COMID, new_POI_ID), by = "member_COMID") %>%
        mutate(POI_ID = ifelse(!is.na(new_POI_ID), new_POI_ID, POI_ID)) %>%
        mutate(aggStep = ifelse(!is.na(new_POI_ID), "coastal", aggStep)) %>%
        select(-new_POI_ID, member_COMID, ID, rpu, POI_ID, aggStep, HUC_12_int, AreaHUC12, intArea) %>%
        mutate(vpu_ID = substr(rpu, 1, 2), comid_char = as.character(member_COMID))
    } else {
      
      divides_poi <- divides_poi %>%
        left_join(st_drop_geometry(divides_miss) %>% select(member_COMID, HUC_12_POI = HUC_12_int), by = "member_COMID") %>%
        mutate(POI_ID = ifelse(!is.na(HUC_12_POI), HUC_12_POI, POI_ID)) %>%
        mutate(aggStep = ifelse(!is.na(HUC_12_POI), "coastal", aggStep)) %>%
        select(-HUC_12_POI, member_COMID, ID, rpu, POI_ID, aggStep, HUC_12_int, AreaHUC12, intArea) %>%
        mutate(vpu_ID = substr(rpu, 1, 2), comid_char = as.character(member_COMID))
    }
    
    divides_poi <- dissolve_holes(divides_poi)
    
    return(divides_poi)
  }
}

#' fix holes
#' @param divides_lu divides layer from create coastal step
#' @param HUC12_table wbd layer from the NHDPlus
#' @param nhdplus_sinks sink layer from the NHDPlus
#' @param min_da_km_terminal minimum drainage area to include a terminal network
#' @param vpu_nhd vpu_nhd from create coastal step
#' @param lookup lookup_table_aggregate from merged vpus
#' @param elev elev from vpu configuration step
#' 
fix_holes <- function(divides_lu, HUC12_table, 
                      nhdplus_sinks, min_da_km_terminal, 
                      vpu_nhd, lookup, elev) {
  
  print(paste0("Currently there are ", sum(is.na(divides_lu$POI_ID)), " cats without a POI_ID"))
  
  if(sum(is.na(divides_lu$POI_ID)) > 0){ 
    
    # Handle nhd sinks
    # arrange divides_lu by member_COMID and populate with RowID for easy iteration within next two steps
    divides_lu <- divides_lu %>%
      arrange(member_COMID, POI_ID) %>%
      mutate(row_ID = row_number()) %>%
      arrange(row_ID)
    
    HUC_sinks <- NHD_sinks(divides_lu, area_thresh = min_da_km_terminal/2,  
                           HUC12_table = HUC12_table, 
                           NHD_sinks = nhdplus_sinks)
    
    if(length(HUC_sinks) == 2){
      divides_lu <- HUC_sinks$divides_poi
      sinks_table <- HUC_sinks$sink_cats_table
    }
    
    # Scan for terminals that may have been refactored
    missing_ds <- filter(divides_lu, is.na(POI_ID))
    
    term_refactored_fun <- function(x) {
      
      ds_ref <- nhdplusTools::get_DM(vpu_nhd, x, include = FALSE)
      
      if(length(ds_ref) == 0){
        return(filter(missing_ds, member_COMID == x))
      }
      
      lookup_test <- filter(lookup, NHDPlusV2_COMID %in% ds_ref)
      
      divides_test <- filter(divides_lu, ID %in% lookup_test$reconciled_ID) %>%
        filter(!is.na(POI_ID))
      
      # print(unique(divides_test$POI_ID))
      
      if(length(unique(divides_test$POI_ID)) == 1){
        return(filter(missing_ds, member_COMID == x) %>%
                 mutate(POI_ID = unique(divides_test$POI_ID)))
      } else {
        return(filter(missing_ds, member_COMID == x)) 
      }
      
    }
    
    term_refactored <- lapply(missing_ds$member_COMID, term_refactored_fun)
    
    term_refactored <- data.table::rbindlist(term_refactored[lengths(term_refactored) > 1],
                                             fill = TRUE) %>%
      st_as_sf()
    
    divides_lu <- filter(divides_lu, !member_COMID %in% term_refactored$member_COMID) %>%
      rbind(term_refactored)
    
    if(sum(is.na(divides_lu$POI_ID)) > 0) {
      
      divides_dem <- miss_term_assign(divides_lu, vpu_nhd, elev) 
      
      divides_lu <- divides_lu %>%
        left_join(select(divides_dem, member_COMID, agg_ID), 
                  by = "member_COMID") %>%
        mutate(POI_ID = ifelse(!is.na(agg_ID), agg_ID, POI_ID),
               aggStep = ifelse(!is.na(agg_ID), "boundary DEM", aggStep)) %>%
        select(-agg_ID)
      
      if(exists("sinks_table")){
        sinks_table_fin <- filter(sinks_table, !member_COMID %in% divides_dem$member_COMID) 
        sinks_table_fin <- data.table::rbindlist(list(sinks_table_fin, 
                                                      divides_dem), fill = TRUE)
      } else {
        sinks_table_fin <- divides_dem
      }
      
      # write_sf(sinks_table_fin, ND_gpkg, ND_table)
    }
    
  } else {
    print ("all unaggregated catchments assigned")
  }
  
  divides_lu <- dissolve_holes(divides_lu)
  
  missing_cats_out <- NULL
  if(sum(is.na(divides_lu$POI_ID)) > 0){
    missing_cats_out <- filter(divides_lu, is.na(POI_ID))
    # write_sf(, ND_gpkg, missing_cats)
  }
  
  # Prob HRU - filter(all_hrus, POI_ID == 140402000209)
  all_hrus <- filter(divides_lu, !is.na(POI_ID)) %>%
    group_by(POI_ID) %>%
    summarize(do_union = TRUE) %>%
    sfheaders::sf_remove_holes(.) %>%
    nhdplusTools:::check_valid(.)
  
  # write_sf(divides_lu, ND_gpkg, divides_nd)
  # write_sf(all_hrus, ND_gpkg, HRU_layer)
  
  list(missing_cats = missing_cats_out, divides_nd = divides_lu, HRU_layer = all_hrus)
}

#' Determines length of shared perimeters with adjacent catchments
#' @param divides (sf data frame) data frame of reconciled divides
#
#' @return (sf data frame) (data frame) DF of shared perimeter lengths for cats
perims <- function(divides_poi){
  
  # Subset to data frame of existing divides with POI assignments
  divides_wpoi <- filter(divides_poi, !is.na(POI_ID)) %>%
    dplyr::select(POI_ID, HUC_12_int, member_COMID, row_ID)
  
  # subset divides with no POI_Id assignment
  divides_nopoi <- filter(divides_poi, is.na(POI_ID)) %>%
    dplyr::select(POI_ID, HUC_12_int, member_COMID, row_ID) #
  
  # Buffer with 1 meter
  divides_nopoi_buff <- st_buffer(divides_nopoi, 1)
  
  divides_nopoi_int <- st_intersection(divides_nopoi_buff, divides_wpoi) %>%
    mutate(perim = lwgeom::st_perimeter(.) / 2) %>%
    dplyr::select(POI_ID, HUC_12_int = HUC_12_int, member_COMID, row_ID, perim, bound_cat_POI_ID = POI_ID.1, 
                  bound_cat_HUC12 = HUC_12_int.1, bound_cat = member_COMID.1, bound_cat_row = row_ID.1) %>%
    st_drop_geometry()
  
  return(divides_nopoi_int)
}

#' Find and lump remaining catchment sinks where possible
#' @param divides_lu (sf data frame) full set of divides and catchments with no assigned POI_ID
#' @param area_thresh (numeric) size threshold for considering sink as its own aggregated unit
#' @param HUC12_table (sf data frame) intersection of divides and HUC12 polygons
#' @param NHD_sinks (sf data frame)  NHD sinks point data frame
#
#' @return (sf data frame) data frame of likely POI_ID assignments of reamining sinks
NHD_sinks <- function(divides_poi, area_thresh, HUC12_table, NHD_sinks){
  
  # populate HUC10 field 
  divides_poi$HUC_10 <- substr(divides_poi$HUC_12_int, 1, 10)
  
  # Filter HUC12 crosswalk to HUC10s id'ed in divides and closed HUC types
  HUC12_sinks <- HUC12_table %>%
    filter(HUC_10 %in% divides_poi$HUC_10) %>%
    filter(HU_10_TYPE == "C" | HU_12_TYPE == "C")
  
  # Determine which HUC12s have only one or less aggregated catchment
  HUC12_poi <- filter(divides_poi, !is.na(POI_ID)) %>%
    distinct(POI_ID, HUC_12_int) %>%
    group_by(HUC_12_int) %>%
    filter(n() < 2, HUC_12_int %in% HUC12_sinks$HUC_12) %>%
    ungroup() %>%
    group_by(POI_ID) %>%
    filter(n() < 2) %>%
    ungroup()
  
  # Get the NHD sinks
  sink_cats_nhd <- filter(divides_poi, is.na(POI_ID)) %>%
    filter(grepl("-", .$member_COMID)) 
  
  # Get the WBD sinks
  sink_cats <- filter(divides_poi, is.na(POI_ID), !member_COMID %in% sink_cats_nhd$member_COMID) %>%
    filter(HUC_10 %in% HUC12_sinks$HUC_10 & HUC_12_int %in% HUC12_sinks$HUC_12) %>%
    rbind(sink_cats_nhd) %>%
    distinct()
  
  # If no sinks in VPU return to main script
  if(nrow(sink_cats) == 0){
    print ("no large sinks in vpu")
    return(divides_poi)
  }
  
  # Check to see if HUC12 alreay identified as POI_ID in HUC12 step
  sink_cats <- sink_cats %>%
    mutate(POI_ID = ifelse(HUC_12_int %in% divides_poi$POI_ID, HUC_12_int, POI_ID))
  
  if(nrow(filter(sink_cats, is.na(POI_ID))) > 0){
    # Aggretate sink cats to HUC12 and add area
    sink_cats_agg <- filter(sink_cats, is.na(POI_ID)) %>%
      group_by(HUC_12_int)  %>%
      summarize(do_union = TRUE) %>%
      mutate(agg_id = row_number())
    
    m_per_km <- 1000
    # Cast to polygon
    agg_cats_poly <- do.call(rbind, lapply(1:nrow(sink_cats_agg),
                                           function(i){st_cast(sink_cats_agg[i,], "POLYGON")})) %>%
      mutate(area = as.numeric(st_area(.))/m_per_km^2)
    
    # Create new 'HRUs' out of larger sinks above the size threshold
    if (nrow(filter(agg_cats_poly, area > area_thresh)) > 0){
      
      # Filter to only large sinks
      agg_cats_large <- filter(agg_cats_poly, area > area_thresh)
      
      # centroids of divides with no assigned POI_ID
      cents <- st_point_on_surface(sink_cats)
      # intersect centroids with the aggregated HUC12 polygons
      cats_int <- st_intersects(cents, agg_cats_large)
      
      # Index of HRUs that missing cats may lie within with holes filled
      agg_cats_index <- sapply(cats_int, function(s) if (length(s) == 0) NA else s)
      
      # Populate ID of HRUs where applicable
      sink_cats$POI_ID_new <- agg_cats_large[agg_cats_index,]$HUC_12_int
      sink_cats <- mutate(sink_cats, aggStep = ifelse(!is.na(POI_ID_new), "HUC12_sinks, large", aggStep))
      sink_cats_final <-  dplyr::select(st_drop_geometry(sink_cats), member_COMID, POI_ID_new) %>%
        filter(!is.na(POI_ID_new))
      
      # Bring over POI_ID to divides
      divides_poi <- divides_poi %>%
        left_join(sink_cats_final, by = "member_COMID") %>%
        mutate(POI_ID = ifelse(!is.na(POI_ID_new), POI_ID_new, POI_ID),
               aggStep = ifelse(!is.na(POI_ID_new), "HUC12_sinks, large", aggStep)) %>%
        dplyr::select(-POI_ID_new)
      
      print ("looking at aggregating smaller sinks")
      
      sink_cats_small <- filter(sink_cats, !member_COMID %in% filter(sink_cats_final, !is.na(POI_ID_new))$member_COMID) %>%
        filter(grepl("-", .$member_COMID)) 
      
      # Get the NHD sinks
      sink_cats_nhd <- filter(divides_poi, is.na(POI_ID)) %>%
        filter(grepl("-", .$member_COMID)) 
      
      agg_cats_small <- filter(agg_cats_poly, area < area_thresh)
      
      if (nrow(sink_cats_small) > 0){
        # Get pertinent boundaries
        cat_bound <- perims(divides_poi) %>%
          filter(!bound_cat %in% sink_cats_small$member_COMID)
        
        # centroids of divides with no assigned POI_ID
        cents <- st_point_on_surface(sink_cats_small)
        # intersect centroids with the aggregated HUC12 polygons
        cats_int <- st_intersects(cents, agg_cats_small)
        
        # Index of HRUs that missing cats may lie within with holes filled
        sink_cats_index <- sapply(cats_int, function(s) if (length(s) == 0) NA else s)
        
        # Populate ID of HRUs where applicable
        sink_cats_small$agg_id <- agg_cats_small[sink_cats_index,]$agg_id
        
        sink_cats_nhd_table <- st_drop_geometry(sink_cats_small) %>%
          inner_join(dplyr::select(cat_bound, -HUC_12_int), by = "member_COMID") %>%
          distinct()
        
        small_sinks_POI <- sink_cats_nhd_table %>%
          distinct(agg_id, HUC_12_int, bound_cat_HUC12, bound_cat_POI_ID) %>%
          #summarize(perim = sum(perim)) %>%
          filter(HUC_12_int == bound_cat_HUC12) %>%
          group_by(agg_id) %>%
          filter(n_distinct(bound_cat_POI_ID) == 1) %>%
          ungroup()
        
        sink_cats_assign <- dplyr::select(sink_cats_nhd_table, member_COMID, agg_id) %>%
          inner_join(dplyr::select(small_sinks_POI, agg_id, POI_ID_new = bound_cat_POI_ID), by = "agg_id") %>%
          distinct()
        
        # Bring over POI_ID to divides
        divides_poi <- divides_poi %>%
          left_join(sink_cats_assign, by = "member_COMID") %>%
          mutate(POI_ID = ifelse(!is.na(POI_ID_new), POI_ID_new, POI_ID),
                 aggStep = ifelse(!is.na(POI_ID_new), "HUC12_sinks, small", aggStep)) %>%
          dplyr::select(-POI_ID_new)
        
        small_sinks <- select(sink_cats_nhd_table, member_COMID, HUC_12 = HUC_12_int, neighbor_COMID = bound_cat,
                              neighbor_HUC12 = bound_cat_HUC12, agg_ID = bound_cat_POI_ID, agg_id, aggStep) %>%
          filter(member_COMID %in% sink_cats_small$member_COMID, !member_COMID %in% sink_cats_final$member_COMID) %>%
          select(-agg_id) %>%
          filter(HUC_12 == neighbor_HUC12) %>%
          mutate(aggStep = "HUC12_sinks, small") %>%
          st_drop_geometry()
        
      }
    }
    large_sinks <- select(sink_cats, member_COMID, HUC_12 = HUC_12_int, POI_ID, aggStep) %>%
      st_drop_geometry()
    
    if(exists("small_sinks")){
      sinks_table <- data.table::rbindlist(list(large_sinks, small_sinks), fill = TRUE) 
    } else {
      sinks_table <- large_sinks
    }
    
    
  }
  return(list(divides_poi = divides_poi, sink_cats_table = sinks_table))
}

#' Find and lump remaining areas where possible
#' @param term_outlets (sf data frame) terminal outlets not assigned POI_ID or joined to reconciled network
#' @param divides_lu (sf data frame) divides dataset with non-dedritic catchments included
#' @param elev character path to elev_cm file
#' @param nclusters (integer)  number of clusters to run over 
#' @param out_file (file) output cache file from parLapply
#
#' @return (sf data frame) data frame of likely POI_ID assignments of reamining sinks
#'  Keep the below link here until publishd
#' Topo data bins: https://www.sciencebase.gov/catalog/item/5f5154ba82ce4c3d12386a02
miss_term_assign <- function(divides_poi, nhd, elev){
  
  # comid <- single comid of a missing catchment (member_COMID in divides_lu)
  #' Find and lump remaining areas where possible
  #' @param comid (integer) comid of single unaggregated terminal outlet
  #' @param divides_lu (sf data frame) divides dataset with non-dedritic catchments included
  #' @param elev character path to elev_cm file
  #
  #' @return (sf data frame) data frame of likely POI_ID assignments of reamining sinks
  assign_func <- function(incomid, divides_poi, nhd, elev) {
    message(incomid)
    
    library(dplyr)
    library(sf)
    library(nhdplusTools)
    
    us_network <- get_UT(nhd, incomid)
    
    # Catchment divide not assigned/aggregated and upstream components
    if (length(us_network) > 0) {
      missing_cats <- filter(divides_poi, member_COMID %in% us_network) 
    } else {
      missing_cats <- filter(divides_poi, member_COMID == incomid)
    }
    
    rpu <- filter(missing_cats, member_COMID == incomid)$rpu
    
    missing_cats_union <- st_union(missing_cats) %>%
      st_as_sf()
    
    # Aggregated catchment Divides
    divides_lu_poi <- filter(divides_poi, !member_COMID %in% us_network)
    
    # Buffered the missing catchment by 1-meter and intersected with
    #     the aggregated catchment divides
    cats_buff <- st_buffer(missing_cats_union, dist = 5) %>%
      st_intersection(divides_lu_poi) %>%
      filter(!member_COMID == incomid)
    
    if(nrow(cats_buff) == 0){
      miss_cat <- dplyr::select(st_drop_geometry(missing_cats), 
                                outlet_COMID = member_COMID) %>%
        dplyr::mutate(outlet_COMID = as.integer(outlet_COMID),
                      Elev = NA, neighbor_COMID = NA, POI_ID = NA) %>%
        dplyr::select(Elev, outlet_COMID, neighbor_COMID, POI_ID)
      return(miss_cat)
    } else {
      # convert cat buf to format for extract input
      buf <- st_as_sf(cats_buff)
      
      #**************************************************************
      # Then do a zonal statitics type operation of the DEM within each 
      #      buffered polygon to get the min cell value, this is a hard coded path to where the dems reside
      
      elev <- as.character(elev)
      elev <- elev[grepl(unique(missing_cats$rpu), elev)]
      
      dem <- terra::rast(elev)
      ex <- terra::extract(dem, terra::vect(buf), min)
      buf_df <- as.data.frame(ex)[,2, drop = FALSE]
      
      # Attribute with values
      buf_df$outlet_COMID <- incomid
      buf_df$neighbor_COMID <- as.character(cats_buff$member_COMID)
      buf_df$POI_ID = as.character(cats_buff$POI_ID)
      
      colnames(buf_df) <- c("Elev","outlet_COMID", "neighbor_COMID", "POI_ID")
      buf_df2 <- buf_df[which.min(buf_df$Elev), ]
      
      return(buf_df2)
    }
  }
  
  term_outlets <- filter(divides_poi, is.na(POI_ID) | aggStep == "HUC12_sinks, small") %>%
    dplyr::rename(outlet_COMID = member_COMID)
  
  out_df <- do.call(rbind,
                    pbapply::pblapply(unique(term_outlets$outlet_COMID),
                                      assign_func, divides_poi, nhd, elev, cl = NULL)) %>%
    rename(member_COMID = outlet_COMID)
  
  # out_df <- lapply(unique(term_outlets$outlet_COMID), 
  #                                    assign_func, divides_poi, nhd, elev)
  
  out_df <- out_df %>%
    left_join(dplyr::select(st_drop_geometry(divides_poi), HUC_12 = HUC_12_int, member_COMID), 
              by = "member_COMID") %>%
    left_join(dplyr::select(st_drop_geometry(divides_poi), neighbor_HUC12 = HUC_12_int, member_COMID), 
              by = c("neighbor_COMID" = "member_COMID")) %>%
    dplyr::select(member_COMID, neighbor_COMID, HUC_12, neighbor_HUC12, agg_ID = POI_ID) %>%
    mutate(aggStep = "boundary DEM")
  
  return(out_df)
}

#' manual checks
#' @param noagg_divides missing_cats from fix holes step
#' @param divides_lu divides from fix holes step
#' @param vpu what vpu we are running
#' 
manual_checks <- function(noagg_divides, divides_lu, vpu) {
  
  noagg_divides <- noagg_divides %>% #read_sf(ND_gpkg, missing_cats) %>%
    select(row_ID, POI_ID_noagg = POI_ID) %>%
    st_drop_geometry()
  
  all_hrus <- NULL
  
  if(all(!is.na(noagg_divides$POI_ID_noagg))){
    print ("all unaggregated divides accounted for")
    
    divides_lu <- divides_lu %>%
      left_join(noagg_divides, by = "row_ID") %>%
      mutate(POI_ID = ifelse(!is.na(POI_ID_noagg), POI_ID_noagg, POI_ID),
             aggStep = ifelse(!is.na(aggStep), "manual", aggStep)) %>%
      select(-POI_ID_noagg)
    
    # Prob HRU - filter(all_hrus, POI_ID == 140402000209)
    all_hrus <- filter(divides_lu, !is.na(POI_ID)) %>%
      group_by(POI_ID) %>%
      summarize(do_union = TRUE) %>%
      sfheaders::sf_remove_holes(.) %>%
      nhdplusTools:::check_valid(.)
    
    # write_sf(divides_lu, ND_gpkg, divides_nd)
    # write_sf(all_hrus, ND_gpkg, HRU_layer) 
    list(HRU_layer = all_hrus, divides_nd = divides_lu)
  } else {
    print ("account for unaggregated divides in ", vpu)
    NULL
  }
  
}