library(targets)

tar_option_set(packages = c("dplyr", "sf", "hyfabric", "hydroloom", "nhdplusTools"),
               memory = "transient", garbage_collection = TRUE, error = "stop", 
               storage = "worker", retrieval = "worker", deployment = "main")

library(future)
library(future.callr)
plan(multisession)

source("R/config.R")
source("R/user_vars.R")
source("R/utils.R")
source("R/02_POI_creation_functions.R")

prep_path <- ".targets_stores/01_prep/objects/"

list(tar_target(data_paths_file, "cache/data_paths.json", format = "file"),
     tar_target(data_paths, jsonlite::read_json(data_paths_file)),
     tar_target(vpu_codes_file, file.path(prep_path, "vpu_codes"), format = "file"),
     tar_target(vpu_codes, readRDS(vpu_codes_file)),
     tar_target(vpu_gpkg, file.path("cache", paste0("reference_", vpu_codes,".gpkg")),
                pattern = map(vpu_codes), deployment = "worker"),
     
     ### Base flowline -- gets updated attributes
     tar_target(flowline, read_sf(vpu_gpkg, nhd_flowline),
                pattern = map(vpu_gpkg), deployment = "worker"),
     # read_sf(data_paths$nhdplus_gdb, "NHDArea")
     tar_target(all_nhdplus_attributes, 
                sf::st_drop_geometry(sf::read_sf(data_paths$nhdplus_gdb, "NHDFlowline_Network"))),
     tar_target(nhdplushr_dir, data_paths$nhdplushr_dir),
     tar_target(full_cat_file, data_paths$fullcats_table, format = "file"),
     tar_target(full_cat_table, readRDS(full_cat_file)),
     
     ### huc12 pois
     tar_target(hu12_points_file, data_paths$hu12_points_path, format = "file"),
     tar_target(huc12_poi, create_hu12_pois(read_sf(hu12_points_file, "hu_points"),
                                       flowline,
                                       poi_name = "HUC12",
                                       vpu_codes),
                pattern = map(flowline, vpu_codes)),
     
     ### gage POIs
     tar_target(gage_info_gpkg_file, gage_info_gpkg, format = "file"),
     tar_target(gages, read_sf(gage_info_gpkg_file, "Gages")),
     tar_target(gage_pois, create_gage_pois(gages,
                                            flowline, "Gages", huc12_poi, 
                                            min_da_km_gages, combine_meters, reach_meas_thresh),
                pattern = map(flowline, huc12_poi), deployment = "worker"),
     tar_target(gage_events, gage_pois$events, pattern = map(gage_pois), deployment = "worker"),
     
     ### Thermo electric power plant POIs
     tar_target(te_points_file, data_paths$TE_points_path, format = "file"),
     tar_target(te_points, read_sf(te_points_file, "2015_TE_Model_Estimates_lat.long_COMIDs")),
     tar_target(te_attributes_file, file.path(data_paths$TE_points_path, "1_model_inputs/te_plants.csv"), format = "file"),
     tar_target(te_attributes, read.csv(te_attributes_file, colClasses = "character")),
     tar_target(te_pois, create_thermo_electric_pois(te_points, te_attributes, 
                                                     gage_pois, huc12_poi$HUC12_COMIDs,
                                                     flowline, "TE"),
                pattern = map(gage_pois, huc12_poi, flowline)),
     
     ### get waterbodies in order
     tar_target(combined_waterbodies_file, file.path(prep_path, "combined_waterbodies"), format = "file"),
     tar_target(combined_waterbodies, readRDS(combined_waterbodies_file)),
     tar_target(nhdplus_area, filter(combined_waterbodies, layer == "NHDArea")),
     tar_target(waterbodies, make_waterbodies_layer(combined_waterbodies, flowline, geo_crs), 
                pattern = map(flowline), deployment = "worker"),

     ### star fit cross walk and reservoir POIs     
     tar_target(istarf_xwalk_file, data_paths$istarf_xwalk, format = "file"),
     tar_target(istarf_xwalk, read.csv(istarf_xwalk_file)),
     tar_target(resops_pois, create_resops_pois(te_pois, waterbodies, istarf_xwalk, 
                                                flowline, "resops"),
                pattern = map(te_pois, waterbodies, flowline), deployment = "worker"),
     tar_target(resops_waterbody_attributes, resops_pois$wb_resops, 
                pattern = map(resops_pois), deployment = "worker"),
     
     ### hilarri is more reservoir POIs
     tar_target(hilarri_file, data_paths$hilarri_sites, format = "file"),
     tar_target(hilarri_data, read.csv(hilarri_file, colClasses = "character")),
     tar_target(hilarri_pois, create_hilarri_pois(resops_pois, 
                                                  hilarri_data, 
                                                  resops_pois$wb_pois,
                                                  flowline, "hilarri"),
                pattern = map(resops_pois, flowline), deployment = "worker"),
     tar_target(hilarri_waterbodies, hilarri_pois$wb_pois, 
                pattern = map(hilarri_pois), deployment = "worker"),
     
     ### waterbody outlet POIs
     tar_target(wb_outlet_pois, create_wb_outlet_pois(hilarri_pois,
                                                      nhdplus_area,
                                                      hilarri_waterbodies,
                                                      resops_waterbody_attributes,
                                                      flowline,
                                                      gage_events,
                                                      proj_crs,
                                                      gages,
                                                      "wb_id",
                                                      nhdplushr_dir),
                pattern = map(hilarri_pois, hilarri_waterbodies, resops_waterbody_attributes, 
                              gage_events,flowline), deployment = "worker"),
     tar_target(gage_and_waterbody_outlet_events, wb_outlet_pois$all_events, 
                pattern = map(wb_outlet_pois), deployment = "worker"),
     
     ### update flowlines for clarity
     tar_target(updated_flowline_wb_outlet, wb_outlet_pois$updated_flowline, 
                pattern = map(wb_outlet_pois), deployment = "worker"),
     
     ### Addition and Removal events from Brakebill, 2019
     tar_target(ar_data, readxl::read_xlsx(file.path(data_paths$USGS_IT_path, 
                                                       "ARevents_final7.xlsx"))),
     tar_target(ar_event_pois, create_ar_event_pois(wb_outlet_pois, ar_data, 
                                                    updated_flowline_wb_outlet, "AR"),
                pattern = map(wb_outlet_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### terminal POIs at network ends
     tar_target(terminal_pois, create_terminal_pois(ar_event_pois, 
                                                    updated_flowline_wb_outlet, 
                                                    min_da_km_terminal, "Term"),
                pattern = map(ar_event_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### confluence POIs
     tar_target(confluence_pois, create_confluence_pois(terminal_pois, 
                                                        updated_flowline_wb_outlet, "Conf"),
                pattern = map(terminal_pois, updated_flowline_wb_outlet), deployment = "worker"),
     
     ### more updated flowline attributes 
     tar_target(updated_flowline_confluence, confluence_pois$updated_flowline, 
                pattern = map(confluence_pois), deployment = "worker"),
     
     ## waterbody inlet POIs
     tar_target(wb_inlet_pois, create_wb_inlet_pois(confluence_pois,
                                                    wb_outlet_pois$WB_VPU_pois,
                                                    gage_and_waterbody_outlet_events,
                                                    updated_flowline_confluence, gages,
                                                    nhdplus_area, proj_crs, "WBIn", nhdplushr_dir),
                pattern = map(confluence_pois, wb_outlet_pois, gage_and_waterbody_outlet_events, 
                              updated_flowline_confluence), deployment = "worker"),
     tar_target(gage_and_all_waterbody_events, wb_inlet_pois$events, 
                pattern = map(wb_inlet_pois), deployment = "worker"),
     
     tar_target(nid_points, read_sf(data_paths$NID_points_path, "Final_NID_2018")),
     tar_target(nid_pois, create_nid_pois(wb_inlet_pois, 
                                          nid_points,
                                          updated_flowline_confluence,
                                          "NID"),
                pattern = map(wb_inlet_pois, updated_flowline_confluence)),
     
     tar_target(headwater_pois, create_headwater_pois(nid_pois, 
                                                      updated_flowline_confluence, 
                                                      all_nhdplus_attributes, 
                                                      min_da_km_hw, "DA"),
                pattern = map(nid_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(elevation_break_pois, create_elevation_break_pois(headwater_pois, 
                                                                  updated_flowline_confluence, 
                                                                  all_nhdplus_attributes,
                                                                  "elev"),
                pattern = map(headwater_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(time_of_travel_pois, create_time_of_travel_pois(elevation_break_pois,
                                                                updated_flowline_confluence,
                                                                all_nhdplus_attributes,
                                                                "Travel"),
                pattern = map(elevation_break_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(final_pois, create_final_pois(time_of_travel_pois,
                                              updated_flowline_confluence),
                pattern = map(time_of_travel_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(draft_segments, create_draft_segments(final_pois, 
                                                      updated_flowline_confluence),
                pattern = map(final_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(collapsed_pois, collapse_pois(final_pois, poi_dar_move, poi_distance_move,
                                              updated_flowline_confluence), 
                pattern = map(final_pois, updated_flowline_confluence), deployment = "worker"),
     
     tar_target(poi_lookup, 
                create_poi_lookup(collapsed_pois, gage_and_all_waterbody_events, 
                                  full_cat_table, updated_flowline_confluence),
                pattern = map(collapsed_pois, gage_and_all_waterbody_events, updated_flowline_confluence)),
     tar_target(write_poi_, {
                  write_sf(poi_lookup$pois_data, vpu_gpkg, poi_data_table)
                  write_sf(poi_lookup$poi_geometry, vpu_gpkg, poi_geometry_table)
                  write_sf(poi_lookup$event_geometry_table, vpu_gpkg, event_geometry_table)
                }, pattern = map(poi_lookup, vpu_gpkg), deployment = "worker")
)