library(targets)

tar_option_set(packages = c("hyRefactor", "dplyr", "sf"),
               memory = "transient", garbage_collection = TRUE, error = "stop", 
               storage = "worker", retrieval = "worker", deployment = "main")

library(future)
library(future.callr)
plan(multisession)

source("R/config.R")
source("R/user_vars.R")
source("R/05_non_dendritic_functions.R")

prep_path <- ".targets_stores/01_prep/objects/"

list(tar_target(data_paths_file, "cache/data_paths.json", format = "file"),
     tar_target(data_paths, jsonlite::read_json(data_paths_file)),
     
     tar_target(all_nhdplus_attributes, 
                sf::st_drop_geometry(sf::read_sf(data_paths$nhdplus_gdb, "NHDFlowline_Network"))),
     tar_target(nhdplus_sinks, read_sf(data_paths$nhdplus_gdb, "Sink")),
     tar_target(nhdplus_wbd, read_sf(data_paths$nhdplus_gdb, "HUC12") |> 
                  st_make_valid() |> st_transform(crs = proj_crs)),
     tar_target(cat_rpu_table, readRDS(data_paths$fullcats_table)),
     
     tar_target(rpu_vpu_file, file.path(prep_path, "rpu_vpu"), format = "file"),
     tar_target(rpu_vpu, readRDS(rpu_vpu_file)),
     tar_target(rpu_vpu_out_file, file.path(prep_path, "rpu_vpu_out"), format = "file"),
     tar_target(rpu_vpu_out, readRDS(rpu_vpu_out_file)),
     
     ## set up to map over vpu_codes
     tar_target(vpu_codes_file, file.path(prep_path, "vpu_codes"), format = "file"),
     tar_target(vpu_codes, readRDS(vpu_codes_file)),
     tar_target(vpu_gpkg, file.path("cache", paste0("reference_", vpu_codes,".gpkg")),
                pattern = map(vpu_codes), deployment = "worker", format = "file"),
     
     # we only write to this gpkg
     tar_target(rfc_gpkg, file.path("cache", paste0("refactor_", vpu_codes,".gpkg")),
                pattern = map(vpu_codes), deployment = "worker"),
     tar_target(gf_gpkg, file.path("cache", paste0("GF_", vpu_codes, ".gpkg")),
                pattern = map(vpu_codes), deployment = "worker"),
     
     # TODO: setup the refactor and aggregate geopackages as format = "file" targets?
     tar_target(merged_vpu, merge_rpu_to_vpu(read_sf(vpu_gpkg,  poi_data_table),
                                             rpu_vpu, vpu_codes, rpu_vpu_out, 
                                             \(rpu) file.path("cache", paste0(rpu, "_refactor.gpkg")),
                                             \(rpu)  file.path("cache", paste0(rpu, "_aggregate.gpkg")),
                                             lookup_table_refactor, reconciled_layer,
                                             divide_layer, split_divide_layer,
                                             agg_fline_layer, agg_cats_layer, 
                                             mapped_outlets_layer),
                pattern = map(vpu_gpkg, vpu_codes), deployment = "worker"),
     
     tar_target(write_out, {
       write_sf(merged_vpu$pois, gf_gpkg, poi_data_table)
       write_sf(merged_vpu$lookup_table_refactor, rfc_gpkg, lookup_table_refactor)
       write_sf(merged_vpu$reconciled_layer, rfc_gpkg, reconciled_layer)
       write_sf(merged_vpu$divide_layer, rfc_gpkg, divide_layer)
       write_sf(merged_vpu$split_divide_layer, rfc_gpkg, split_divide_layer)
       write_sf(merged_vpu$mapped_outlets_layer, gf_gpkg, mapped_outlets_layer)
       write_sf(merged_vpu$agg_cats_layer, gf_gpkg, agg_cats_layer)
       write_sf(merged_vpu$agg_fline_layer, gf_gpkg, agg_fline_layer)
       write_sf(merged_vpu$lookup_table_aggregate, gf_gpkg, lookup_table_refactor)
       write_sf(merged_vpu$catchment_network_table_refector, rfc_gpkg, catchment_network_table)
       write_sf(merged_vpu$cathment_network_table_aggregate, gf_gpkg, catchment_network_table)
     }, pattern = map(merged_vpu, gf_gpkg, rfc_gpkg), deployment = "worker"),
     
     ## prep for non-dendritic
     tar_target(vpu_attributes, get_vpu_config(vpu_codes, data_paths, 
                                               all_nhdplus_attributes, nhdplus_wbd),
                pattern = map(vpu_codes)),
     # NOTUSED?
     # tar_target(flowline_vpu, st_transform(read_sf(vpu_gpkg, nhd_flowline), proj_crs),
     #            pattern = map(vpu_gpkg), deployment = "worker"),
     tar_target(catchment_vpu, st_transform(read_sf(vpu_gpkg, nhd_catchment), proj_crs),
                pattern = map(vpu_gpkg), deployment = "worker"),
     tar_target(divides_vpu, st_transform(merged_vpu$divide_layer, proj_crs),
                pattern = map(merged_vpu), deployment = "worker"),
     # NOTUSED?
     # tar_target(lookup_vpu, merged_vpu$lookup_table_aggregate,
     #            pattern = map(merged_vpu), deployment = "worker"),
     
     tar_target(hu12_xwalk, make_hu12_xwalk(catchment_vpu, divides_vpu, vpu_codes, 
                                            # TODO track down this file or recreate it!!
                                            file.path(data_paths$nhdplus_dir,
                                                      "CrosswalkTable_NHDplus_HU12.csv"),
                                            nhdplus_wbd),
                pattern = map(catchment_vpu, divides_vpu, vpu_codes), deployment = "worker"),
     tar_target(divides_non_dendritic, make_divides_nd(hu12_xwalk$divide_layer, 
                                                       merged_vpu$lookup_table_aggregate, 
                                                       hu12_xwalk$nhd_catchment,
                                                       cat_rpu_table,
                                                       debug = TRUE),
                pattern = map(hu12_xwalk, merged_vpu), deployment = "worker"),
     tar_target(divides_w_huc12, assign_huc12_fun(divides_non_dendritic$divides_nd, 
                                                  hu12_xwalk$xwalk_layer, 
                                                  vpu_attributes$vpu_nhd, 
                                                  debug = FALSE),
                pattern = map(divides_non_dendritic, hu12_xwalk, vpu_attributes), deployment = "worker"),
     tar_target(divides_w_coastal, create_coastal_catchments(vpu_attributes$vpu_nhd, 
                                                             divides_w_huc12$divides_nd,
                                                             vpu_attributes$vpu_WBD, 
                                                             debug = FALSE),
                pattern = map(vpu_attributes, divides_w_huc12), deployment = "worker"),
     tar_target(divides_fixed_holes, fix_holes(divides_w_coastal$divides_nd, 
                                               nhdplus_wbd, nhdplus_sinks, 
                                               min_da_km_terminal,
                                               divides_w_coastal$vpu_nhd,
                                               merged_vpu$lookup_table_aggregate,
                                               vpu_attributes$elev),
                pattern = map(divides_w_coastal, vpu_attributes, merged_vpu), deployment = "worker"),
     tar_target(final, manual_checks(divides_fixed_holes$missing_cats, divides_fixed_holes$divides_nd, vpu_codes),
                pattern = map(divides_fixed_holes, vpu_codes), deployment = "worker")
)
